#!/usr/bin/env python
""" File Catalog Client Command Line Interface. """

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

# TODO: This should be modernised to use subprocess(32)
try:
    import commands
except ImportError:
    # Python 3's subprocess module contains a compatibility layer
    import subprocess as commands
import os.path
import time
import sys
import getopt

from DIRAC import S_ERROR, S_OK
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Base.CLI import CLI
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Core.Utilities.PrettyPrint import int_with_commas, printTable
from DIRAC.DataManagementSystem.Client.DirectoryListing import DirectoryListing
from DIRAC.DataManagementSystem.Client.MetaQuery import MetaQuery, FILE_STANDARD_METAKEYS
from DIRAC.DataManagementSystem.Client.CmdDirCompletion.AbstractFileSystem import DFCFileSystem, UnixLikeFileSystem
from DIRAC.DataManagementSystem.Client.CmdDirCompletion.DirectoryCompletion import DirectoryCompletion

__RCSID__ = "$Id$"


class FileCatalogClientCLI(CLI):
    """usage: FileCatalogClientCLI.py xmlrpc-url.

    The URL should use HTTP protocol, and specify a port.  e.g.::

        http://localhost:7777

    This provides a command line interface to the FileCatalog Exported API::

        ls(path) - lists the directory path

    The command line interface to these functions can be listed by typing "help"
    at the prompt.

    Other modules which want access to the FileCatalog API should simply make
    their own internal connection to the XMLRPC server using code like::

        server = xmlrpclib.Server(xmlrpc_url)
        server.exported_function(args)
    """

    def __init__(self, client):
        CLI.__init__(self)
        self.fc = client
        self.cwd = "/"
        self.prompt = "FC:" + self.cwd + "> "
        self.previous_cwd = "/"

        self.dfc_fs = DFCFileSystem(self.fc)
        self.lfn_dc = DirectoryCompletion(self.dfc_fs)

        self.ul_fs = UnixLikeFileSystem()
        self.ul_dc = DirectoryCompletion(self.ul_fs)

    def getPath(self, apath):

        if apath.find("/") == 0:
            path = apath
        else:
            path = self.cwd + "/" + apath
        path = path.replace("//", "/")

        return os.path.normpath(path)

    def do_register(self, args):
        """Register a record to the File Catalog

        usage:
          register file <lfn> <pfn> <size> <SE> <guid> <Checksum>  - register new file record in the catalog
          register replica <lfn> <pfn> <SE>   - register new replica in the catalog
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_register.__doc__)
            return
        option = argss[0]
        del argss[0]
        if option == "file":
            if len(argss) < 4:
                print(self.do_register.__doc__)
                return
            return self.registerFile(argss)
        elif option == "pfn" or option == "replica":
            # TODO
            # Is the __doc__ not complete ?
            if len(argss) != 3:
                print(self.do_register.__doc__)
                return
            return self.registerReplica(argss)
        else:
            print("Unknown option:", option)

    # An Auto Completion For ``register``
    _available_register_cmd = ["file", "replica"]

    def complete_register(self, text, line, begidx, endidx):
        result = []
        args = line.split()
        if len(args) >= 2 and (args[1] in self._available_register_cmd):
            # if 'register file' or 'register replica' exists,
            # try to do LFN auto completion.
            cur_path = ""
            if len(args) == 3:
                cur_path = args[2]
            result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)
            return result

        result = [i for i in self._available_register_cmd if i.startswith(text)]
        return result

    def do_add(self, args):
        """Upload a new file to a SE and register in the File Catalog

        usage:

          add <lfn> <pfn> <SE> [<guid>]
        """

        # ToDo - adding directories

        argss = args.split()

        if len(argss) < 3:
            print("Error: insufficient number of arguments")
            return

        lfn = argss[0]
        lfn = self.getPath(lfn)
        pfn = argss[1]
        se = argss[2]
        guid = None
        if len(argss) > 3:
            guid = argss[3]

        dirac = Dirac()
        result = dirac.addFile(lfn, pfn, se, guid, printOutput=False)
        if not result["OK"]:
            print("Error: %s" % (result["Message"]))
        else:
            print("File %s successfully uploaded to the %s SE" % (lfn, se))

    def complete_add(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_get(self, args):
        """Download file from grid and store in a local directory

        usage:

          get <lfn> [<local_directory>]
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_get.__doc__)
            return
        lfn = argss[0]
        lfn = self.getPath(lfn)
        dir_ = ""
        if len(argss) > 1:
            dir_ = argss[1]

        dirac = Dirac()
        localCWD = ""
        if dir_:
            localCWD = os.getcwd()
            os.chdir(dir_)
        result = dirac.getFile(lfn)
        if localCWD:
            os.chdir(localCWD)

        if not result["OK"]:
            print("Error: %s" % (result["Message"]))
        else:
            print("File %s successfully downloaded" % lfn)

    def complete_get(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_unregister(self, args):
        """Unregister records in the File Catalog

        usage:
          unregister replica  <lfn> <se>
          unregister file <lfn>
          unregister dir <path>
        """
        argss = args.split()
        if len(argss) == 0:
            print(self.do_unregister.__doc__)
            return
        option = argss[0]
        del argss[0]
        if option == "replica":
            if len(argss) != 2:
                print(self.do_unregister.__doc__)
                return
            return self.removeReplica(argss)
        elif option == "file":
            if len(argss) != 1:
                print(self.do_unregister.__doc__)
                return
            return self.removeFile(argss)
        elif option == "dir" or option == "directory":
            if len(argss) != 1:
                print(self.do_unregister.__doc__)
                return
            return self.removeDirectory(argss)
        else:
            print("Error: illegal option %s" % option)

    # An Auto Completion For ``register``
    _available_unregister_cmd = ["replica", "file", "dir", "directory"]

    def complete_unregister(self, text, line, begidx, endidx):
        result = []
        args = line.split()
        if len(args) >= 2 and (args[1] in self._available_unregister_cmd):
            # if 'unregister file' or 'unregister replica' and so on exists,
            # try to do LFN auto completion.
            cur_path = ""
            if len(args) == 3:
                cur_path = args[2]
            result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)
            return result

        result = [i for i in self._available_unregister_cmd if i.startswith(text)]
        return result

    def do_rmreplica(self, args):
        """Remove LFN replica from the storage and from the File Catalog

        usage:
          rmreplica <lfn> <se>
        """
        argss = args.split()
        if len(argss) != 2:
            print(self.do_rmreplica.__doc__)
            return
        lfn = argss[0]
        lfn = self.getPath(lfn)
        print("lfn:", lfn)
        se = argss[1]
        try:
            result = self.fc.setReplicaStatus({lfn: {"SE": se, "Status": "Trash"}})
            if result["OK"]:
                print("Replica at", se, "moved to Trash Bin")
            else:
                print("Failed to remove replica at", se)
                print(result["Message"])
        except Exception as x:
            print("Error: rmreplica failed with exception: ", x)

    def complete_rmreplica(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_rm(self, args):
        """Remove file from the storage and from the File Catalog

        usage:
          rm <lfn>

        NB: this method is not fully implemented !
        """
        # Not yet really implemented
        argss = args.split()
        if len(argss) != 1:
            print(self.do_rm.__doc__)
            return
        self.removeFile(argss)

    def complete_rm(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_rmdir(self, args):
        """Remove directory from the File Catalog. Note, this method does not remove physical replicas

        usage:
          rmdir [-r] [-f] <path>

        -r flag to remove directories recursively
        -f flag to force removing files also in the directories to be removed

        """
        argss = args.split()
        recursive = False
        if "-r" in argss:
            recursive = True
            argss.remove("-r")
        forceNonEmpty = False
        if "-f" in argss:
            forceNonEmpty = True
            argss.remove("-f")
        if "-rf" in argss:
            recursive = True
            forceNonEmpty = True
            argss.remove("-rf")
        if len(argss) != 1:
            print(self.do_rmdir.__doc__)
            return
        path = argss[0]
        lfn = self.getPath(path)
        self.removeDirectory(lfn, recursive, forceNonEmpty)

    def complete_rmdir(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def removeReplica(self, args):
        """Remove replica from the catalog"""

        path = args[0]
        lfn = self.getPath(path)
        print("lfn:", lfn)
        rmse = args[1]
        try:
            result = self.fc.removeReplica({lfn: {"SE": rmse}})
            if result["OK"]:
                if "Failed" in result["Value"]:
                    if lfn in result["Value"]["Failed"]:
                        print("ERROR: %s" % (result["Value"]["Failed"][lfn]))
                    elif lfn in result["Value"]["Successful"]:
                        print("File %s at %s removed from the catalog" % (lfn, rmse))
                    else:
                        print("ERROR: Unexpected returned value %s" % result["Value"])
                else:
                    print("File %s at %s removed from the catalog" % (lfn, rmse))
            else:
                print("Failed to remove replica at", rmse)
                print(result["Message"])
        except Exception as x:
            print("Error: rmpfn failed with exception: ", x)

    def removeFile(self, args):
        """Remove file from the catalog"""

        path = args[0]
        lfn = self.getPath(path)
        print("lfn:", lfn)
        try:
            result = self.fc.removeFile(lfn)
            if result["OK"]:
                if "Failed" in result["Value"]:
                    if lfn in result["Value"]["Failed"]:
                        print("ERROR: %s" % (result["Value"]["Failed"][lfn]))
                    elif lfn in result["Value"]["Successful"]:
                        print("File", lfn, "removed from the catalog")
                    else:
                        print("ERROR: Unexpected result %s" % result["Value"])
                else:
                    print("File", lfn, "removed from the catalog")
            else:
                print("Failed to remove file from the catalog")
                print(result["Message"])
        except Exception as x:
            print("Error: rm failed with exception: ", x)

    def removeDirectory(self, lfn, recursive=False, forceNonEmpty=False):
        """Remove a given directory from the catalog. Remove multiple directories
        recursively if recursive flag is True. Remove contained files if forceNonEmpty
        flag is true.
        """

        if recursive or forceNonEmpty:
            resultListDirectory = self.fc.listDirectory(lfn, False)
            if not resultListDirectory["OK"]:
                print("Failed to look up the directory contents")
                return S_ERROR("Failed to look up the directory contents")

            # Remove subdirectories first
            dirDict = resultListDirectory["Value"]["Successful"][lfn]["SubDirs"]
            if dirDict:
                if recursive:
                    for dirLfn in dirDict:
                        result = self.removeDirectory(dirLfn, recursive, forceNonEmpty)
                        if not result["OK"]:
                            print("Error: failed to remove directory", dirLfn)
                            return S_ERROR("Failed to remove directory")
                else:
                    print("Error: failed to remove non empty directory")
                    return S_ERROR("Failed to remove non empty directory")

            # Remove files
            fileDict = resultListDirectory["Value"]["Successful"][lfn]["Files"]
            fileList = list(fileDict)
            if fileList:
                if forceNonEmpty:
                    print("Removing", len(fileList), "files in", lfn)
                    result = self.fc.removeFile(fileList)
                    if not result["OK"]:
                        print("Error:", result["Message"])
                        return result
                    if result["Value"]["Failed"]:
                        print("Error: failed to remove %d files" % len(result["Value"]["Failed"]))
                        return S_ERROR("Failed to remove files")
                else:
                    print("Error: failed to remove non empty directory")
                    return S_ERROR("Failed to remove non empty directory")

        # Removing the directory now
        try:
            print("Removing directory", lfn)
            result = returnSingleResult(self.fc.removeDirectory(lfn))
            if not result["OK"]:
                print("Failed to remove directory from the catalog")
                print(result["Message"])
            return result
        except Exception as x:
            print("Error: rmdir failed with exception: ", x)
            return S_ERROR("Exception: %s" % str(x))

    def do_replicate(self, args):
        """Replicate a given file to a given SE

        usage:
          replicate <LFN> <SE> [<SourceSE>]
        """
        argss = args.split()
        if len(argss) < 2:
            print("Error: insufficient number of arguments")
            return
        lfn = argss[0]
        lfn = self.getPath(lfn)
        se = argss[1]
        sourceSE = ""
        if len(argss) > 2:
            sourceSE = argss[2]
        try:
            dirac = Dirac()
            result = dirac.replicateFile(lfn, se, sourceSE, printOutput=True)
            if not result["OK"]:
                print("Error: %s" % (result["Message"]))
            elif not result["Value"]:
                print("Replica is already present at the target SE")
            else:
                print("File %s successfully replicated to the %s SE" % (lfn, se))
        except Exception as x:
            print("Error: replicate failed with exception: ", x)

    def complete_replicate(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_replicas(self, args):
        """Get replicas for the given file specified by its LFN

        usage: replicas <lfn>
        """
        argss = args.split()
        if len(argss) == 0:
            print(self.do_replicas.__doc__)
            return
        apath = argss[0]
        path = self.getPath(apath)
        print("lfn:", path)
        try:
            result = self.fc.getReplicas(path)
            if result["OK"]:
                if result["Value"]["Successful"]:
                    for se, entry in result["Value"]["Successful"][path].items():
                        print(se.ljust(15), entry)
            else:
                print("Replicas: ", result["Message"])
        except Exception as x:
            print("replicas failed: ", x)

    def complete_replicas(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def registerFile(self, args):
        """Add a file to the catalog

        usage: add <lfn> <pfn> <size> <SE> <guid> <Checksum>

        * If pfn value is None, it will be assigned an empty string
        * If guid value is None, it will be generated
        * If Checksum value is None, it will be assigned an empty string
        """

        if len(args) != 6:
            print("Command takes 6 arguments, %d given" % len(args))
            print(self.do_register.__doc__)
            return

        path = args[0]
        infoDict = {}
        lfn = self.getPath(path)
        infoDict["PFN"] = args[1] if args[1] != "None" else ""
        infoDict["Size"] = int(args[2])
        infoDict["SE"] = args[3]
        if args[4] == "None":
            _status, guid = commands.getstatusoutput("uuidgen")
        else:
            guid = args[4]
        infoDict["GUID"] = guid
        infoDict["Checksum"] = args[5] if args[5] != "None" else ""

        fileDict = {}
        fileDict[lfn] = infoDict

        try:
            result = self.fc.addFile(fileDict)
            if not result["OK"]:
                print("Failed to add file to the catalog: ", end=" ")
                print(result["Message"])
            elif result["Value"]["Failed"]:
                if lfn in result["Value"]["Failed"]:
                    print("Failed to add file:", result["Value"]["Failed"][lfn])
            elif result["Value"]["Successful"]:
                if lfn in result["Value"]["Successful"]:
                    print("File successfully added to the catalog")
        except Exception as x:
            print("add file failed: ", str(x))

    def registerReplica(self, args):
        """Add a file to the catatlog

        usage: addpfn <lfn> <pfn> <SE>
        """
        path = args[0]
        infoDict = {}
        lfn = self.getPath(path)
        infoDict["PFN"] = args[1]
        if infoDict["PFN"] == "''" or infoDict["PFN"] == '""':
            infoDict["PFN"] = ""
        infoDict["SE"] = args[2]

        repDict = {}
        repDict[lfn] = infoDict

        try:
            result = self.fc.addReplica(repDict)
            if not result["OK"]:
                print("Failed to add replica to the catalog: ", end=" ")
                print(result["Message"])
            elif result["Value"]["Failed"]:
                print("Failed to add replica:", result["Value"]["Failed"][lfn])
            else:
                print("Replica added successfully:", result["Value"]["Successful"][lfn])
        except Exception as x:
            print("add pfn failed: ", str(x))

    def do_ancestorset(self, args):
        """Set ancestors for the given file

        usage: ancestorset <lfn> <ancestor_lfn> [<ancestor_lfn>...]
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_ancestorset.__doc__)
            return
        lfn = argss[0]
        if lfn[0] != "/":
            lfn = self.cwd + "/" + lfn
        ancestors = argss[1:]
        tmpList = []
        for a in ancestors:
            if a[0] != "/":
                a = self.cwd + "/" + a
            tmpList.append(a)
        ancestors = tmpList

        try:
            result = self.fc.addFileAncestors({lfn: {"Ancestors": ancestors}})
            if not result["OK"]:
                print("Failed to add file ancestors to the catalog: ", end=" ")
                print(result["Message"])
            elif result["Value"]["Failed"]:
                print("Failed to add file ancestors to the catalog: ", end=" ")
                print(result["Value"]["Failed"][lfn])
            else:
                print("Added %d ancestors to file %s" % (len(ancestors), lfn))
        except Exception as x:
            print("Exception while adding ancestors: ", str(x))

    def complete_ancestorset(self, text, line, begidx, endidx):

        args = line.split()

        if len(args) == 1:
            cur_path = ""
        elif len(args) > 1:
            # If the line ends with ' '
            # this means a new parameter begin.
            if line.endswith(" "):
                cur_path = ""
            else:
                cur_path = args[-1]

        result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_ancestor(self, args):
        """Get ancestors of the given file

        usage: ancestor <lfn> [depth]
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_ancestor.__doc__)
            return
        lfn = argss[0]
        if lfn[0] != "/":
            lfn = self.cwd + "/" + lfn
        depth = [1]
        if len(argss) > 1:
            depth = int(argss[1])
            depth = list(range(1, depth + 1))

        try:
            result = self.fc.getFileAncestors([lfn], depth)
            if not result["OK"]:
                print("ERROR: Failed to get ancestors: ", end=" ")
                print(result["Message"])
            elif result["Value"]["Failed"]:
                print("Failed to get ancestors: ", end=" ")
                print(result["Value"]["Failed"][lfn])
            else:
                depthDict = {}
                depSet = set()
                for lfn, ancestorDict in result["Value"]["Successful"].items():
                    for ancestor, dep in ancestorDict.items():
                        depthDict.setdefault(dep, [])
                        depthDict[dep].append(ancestor)
                        depSet.add(dep)
                depList = sorted(depSet)
                print(lfn)
                for dep in depList:
                    for lfn in depthDict[dep]:
                        print(dep, " " * dep * 5, lfn)
        except Exception as x:
            print("Exception while getting ancestors: ", str(x))

    def complete_ancestor(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_descendent(self, args):
        """Get descendents of the given file

        usage: descendent <lfn> [depth]
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_descendent.__doc__)
            return
        lfn = argss[0]
        if lfn[0] != "/":
            lfn = self.cwd + "/" + lfn
        depth = [1]
        if len(argss) > 1:
            depth = int(argss[1])
            depth = list(range(1, depth + 1))

        try:
            result = self.fc.getFileDescendents([lfn], depth)
            if not result["OK"]:
                print("ERROR: Failed to get descendents: ", end=" ")
                print(result["Message"])
            elif result["Value"]["Failed"]:
                print("Failed to get descendents: ", end=" ")
                print(result["Value"]["Failed"][lfn])
            else:
                depthDict = {}
                depSet = set()
                for lfn, descDict in result["Value"]["Successful"].items():
                    for desc, dep in descDict.items():
                        depthDict.setdefault(dep, [])
                        depthDict[dep].append(desc)
                        depSet.add(dep)
                depList = sorted(depSet)
                print(lfn)
                for dep in depList:
                    for lfn in depthDict[dep]:
                        print(dep, " " * dep * 5, lfn)
        except Exception as x:
            print("Exception while getting descendents: ", str(x))

    def complete_descendent(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    #######################################################################################
    # User and group methods

    def do_user(self, args):
        """User related commands

        usage:
          user add <username>  - register new user in the catalog
          user delete <username>  - delete user from the catalog
          user show - show all users registered in the catalog
        """
        argss = args.split()
        if len(argss) == 0:
            print(self.do_user.__doc__)
            return
        option = argss[0]
        del argss[0]
        if option == "add":
            if len(argss) != 1:
                print(self.do_user.__doc__)
                return
            return self.registerUser(argss)
        elif option == "delete":
            if len(argss) != 1:
                print(self.do_user.__doc__)
                return
            return self.deleteUser(argss)
        elif option == "show":
            result = self.fc.getUsers()
            if not result["OK"]:
                print(("Error: %s" % result["Message"]))
            else:
                if not result["Value"]:
                    print("No entries found")
                else:
                    for user, id_ in result["Value"].items():
                        print(user.rjust(20), ":", id_)
        else:
            print("Unknown option:", option)

    # completion for ``user``
    _available_user_cmd = ["add", "delete", "show"]

    def complete_user(self, text, line, begidx, endidx):
        result = []
        args = line.split()
        if len(args) == 2 and (args[1] in self._available_user_cmd):
            # if the sub command exists,
            # Don't need any auto completion
            return result

        result = [i for i in self._available_user_cmd if i.startswith(text)]
        return result

    def do_group(self, args):
        """Group related commands

        usage:
          group add <groupname>  - register new group in the catalog
          group delete <groupname>  - delete group from the catalog
          group show - how all groups registered in the catalog
        """
        argss = args.split()
        if len(argss) == 0:
            print(self.do_group.__doc__)
            return
        option = argss[0]
        del argss[0]
        if option == "add":
            if len(argss) != 1:
                print(self.do_group.__doc__)
                return
            return self.registerGroup(argss)
        elif option == "delete":
            if len(argss) != 1:
                print(self.do_group.__doc__)
                return
            return self.deleteGroup(argss)
        elif option == "show":
            result = self.fc.getGroups()
            if not result["OK"]:
                print(("Error: %s" % result["Message"]))
            else:
                if not result["Value"]:
                    print("No entries found")
                else:
                    for user, id_ in result["Value"].items():
                        print(user.rjust(20), ":", id_)
        else:
            print("Unknown option:", option)

    # completion for ``group``
    _available_group_cmd = ["add", "delete", "show"]

    def complete_group(self, text, line, begidx, endidx):
        result = []
        args = line.split()
        if len(args) == 2 and (args[1] in self._available_group_cmd):
            # if the sub command exists,
            # Don't need any auto completion
            return result

        result = [i for i in self._available_group_cmd if i.startswith(text)]
        return result

    def registerUser(self, argss):
        """Add new user to the File Catalog

        usage: adduser <user_name>
        """

        username = argss[0]

        result = self.fc.addUser(username)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
        else:
            print("User ID:", result["Value"])

    def deleteUser(self, args):
        """Delete user from the File Catalog

        usage: deleteuser <user_name>
        """

        username = args[0]

        result = self.fc.deleteUser(username)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))

    def registerGroup(self, argss):
        """Add new group to the File Catalog

        usage: addgroup <group_name>
        """

        gname = argss[0]

        result = self.fc.addGroup(gname)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
        else:
            print("Group ID:", result["Value"])

    def deleteGroup(self, args):
        """Delete group from the File Catalog

        usage: deletegroup <group_name>
        """

        gname = args[0]

        result = self.fc.deleteGroup(gname)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))

    def do_mkdir(self, args):
        """Make directory

        usage: mkdir <path>
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_mkdir.__doc__)
            return
        path = argss[0]
        if path.find("/") == 0:
            newdir = path
        else:
            newdir = self.cwd + "/" + path

        newdir = self.getPath(newdir)

        result = self.fc.createDirectory(newdir)
        if result["OK"]:
            if result["Value"]["Successful"]:
                if newdir in result["Value"]["Successful"]:
                    print("Successfully created directory:", newdir)
            elif result["Value"]["Failed"]:
                if newdir in result["Value"]["Failed"]:
                    print("Failed to create directory:", result["Value"]["Failed"][newdir])
        else:
            print("Failed to create directory:", result["Message"])

    def complete_mkdir(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_cd(self, args):
        """Change directory to <path>

        usage: cd <path>
               cd -
        """

        argss = args.split()
        if len(argss) == 0:
            path = "/"
        else:
            path = argss[0]

        if path == "-":
            path = self.previous_cwd

        newcwd = self.getPath(path)
        if len(newcwd) > 1 and not newcwd.find("..") == 0:
            newcwd = newcwd.rstrip("/")

        result = self.fc.isDirectory(newcwd)
        if result["OK"]:
            if result["Value"]["Successful"]:
                if result["Value"]["Successful"][newcwd]:
                    # if result['Type'] == "Directory":
                    self.previous_cwd = self.cwd
                    self.cwd = newcwd
                    self.prompt = "FC:" + self.cwd + ">"
                else:
                    print(newcwd, "does not exist or is not a directory")
            else:
                print(newcwd, "is not found")
        else:
            print("Server failed to find the directory", newcwd)

    def complete_cd(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_id(self, args):
        """Get user identity"""
        result = getProxyInfo()
        if not result["OK"]:
            print("Error: %s" % result["Message"])
            return
        user = result["Value"]["username"]
        group = result["Value"]["group"]
        result = self.fc.getUsers()
        if not result["OK"]:
            print("Error: %s" % result["Message"])
            return
        userDict = result["Value"]
        result = self.fc.getGroups()
        if not result["OK"]:
            print("Error: %s" % result["Message"])
            return
        groupDict = result["Value"]
        idUser = userDict.get(user, 0)
        idGroup = groupDict.get(group, 0)
        print("user=%d(%s) group=%d(%s)" % (idUser, user, idGroup, group))

    def do_lcd(self, args):
        """Change local directory

        usage:
          lcd <local_directory>
        """
        argss = args.split()
        if len(argss) != 1:
            print(self.do_lcd.__doc__)
            return
        localDir = argss[0]
        try:
            os.chdir(localDir)
            newDir = os.getcwd()
            print("Local directory: %s" % newDir)
        except Exception:
            print("%s seems not a directory" % localDir)

    def complete_lcd(self, text, line, begidx, endidx):
        # TODO
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.ul_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_pwd(self, args):
        """Print out the current directory

        usage: pwd
        """
        print(self.cwd)

    def do_ls(self, args):
        """Lists directory entries at <path>

           usage: ls [-ltrnSh] <path>

        -l  --long                : Long listing.
        -t  --timeorder           : List ordering by time.
        -r  --reverse             : Reverse list order.
        -n  --numericid           : List with numeric value of UID and GID.
        -S  --sizeorder           : List ordering by file size.
        -H  --human-readable      : Print sizes in human readable format (e.g., 1Ki, 20Mi);
                                    powers of 2 are used (1Mi = 2^20 B).
        """

        argss = args.split()
        # Get switches
        _long = False
        reverse = False
        timeorder = False
        numericid = False
        sizeorder = False
        humanread = False
        shortopts = "ltrnSH"
        longopts = ["long", "timeorder", "reverse", "numericid", "sizeorder", "human-readable"]
        path = self.cwd
        if len(argss) > 0:
            try:
                optlist, arguments = getopt.getopt(argss, shortopts, longopts)
            except getopt.GetoptError as e:
                print(str(e))
                print(self.do_ls.__doc__)
                return
            # Duplicated options are allowed: later options have precedence, e.g.,
            # '-ltSt' will be order by time
            # '-ltStS' will be order by size
            options = [opt for (opt, arg) in optlist]
            for opt in options:
                if opt in ["-l", "--long"]:
                    _long = True
                elif opt in ["-r", "--reverse"]:
                    reverse = True
                elif opt in ["-t", "--timeorder"]:
                    timeorder = True
                elif opt in ["-n", "--numericid"]:
                    numericid = True
                elif opt in ["-S", "--sizeorder"]:
                    sizeorder = True
                elif opt in ["-H", "--human-readable"]:
                    humanread = True

            if timeorder and sizeorder:
                options = [w.replace("--sizeorder", "-S") for w in options]
                options = [w.replace("--human-readable", "-H") for w in options]
                options.reverse()
                # The last ['-S','-t'] provided is the one we use: reverse order
                # means that the last provided has the smallest index.
                if options.index("-S") < options.index("-t"):
                    timeorder = False
                else:
                    sizeorder = False

            # Get path
            if arguments:
                inputpath = False
                while arguments or not inputpath:
                    tmparg = arguments.pop()
                    # look for a non recognized option not starting with '-'
                    if tmparg[0] != "-":
                        path = tmparg
                        inputpath = True
                        if path[0] != "/":
                            path = self.cwd + "/" + path
        path = self.getPath(path)

        # Check if the target path is a file
        result = self.fc.isFile(path)
        if not result["OK"]:
            print("Error: can not verify path")
            return
        elif path in result["Value"]["Successful"] and result["Value"]["Successful"][path]:
            result = self.fc.getFileMetadata(path)
            dList = DirectoryListing()
            fileDict = result["Value"]["Successful"][path]
            dList.addFile(os.path.basename(path), fileDict, {}, numericid)
            dList.printListing(reverse, timeorder, sizeorder, humanread)
            return

        # Get directory contents now
        try:
            result = self.fc.listDirectory(path, _long)
            dList = DirectoryListing()
            if result["OK"]:
                if result["Value"]["Successful"]:
                    for entry in result["Value"]["Successful"][path]["Files"]:
                        fname = entry.split("/")[-1]
                        # print entry, fname
                        # fname = entry.replace(self.cwd,'').replace('/','')
                        if _long:
                            fileDict = result["Value"]["Successful"][path]["Files"][entry]["MetaData"]
                            repDict = result["Value"]["Successful"][path]["Files"][entry].get("Replicas", {})
                            if fileDict:
                                dList.addFile(fname, fileDict, repDict, numericid)
                        else:
                            dList.addSimpleFile(fname)
                    for entry in result["Value"]["Successful"][path]["SubDirs"]:
                        dname = entry.split("/")[-1]
                        # print entry, dname
                        # dname = entry.replace(self.cwd,'').replace('/','')
                        if _long:
                            dirDict = result["Value"]["Successful"][path]["SubDirs"][entry]
                            if dirDict:
                                dList.addDirectory(dname, dirDict, numericid)
                        else:
                            dList.addSimpleFile(dname)

                    for entry in result["Value"]["Successful"][path]["Links"]:
                        pass

                    if "Datasets" in result["Value"]["Successful"][path]:
                        for entry in result["Value"]["Successful"][path]["Datasets"]:
                            dname = os.path.basename(entry)
                            if _long:
                                dsDict = result["Value"]["Successful"][path]["Datasets"][entry]["Metadata"]
                                if dsDict:
                                    dList.addDataset(dname, dsDict, numericid)
                            else:
                                dList.addSimpleFile(dname)

                    if _long:
                        dList.printListing(reverse, timeorder, sizeorder, humanread)
                    else:
                        dList.printOrdered()
            else:
                print("Error:", result["Message"])
        except Exception as x:
            print("Error:", str(x))

    def complete_ls(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        index_cnt = 0

        if len(args) > 1:
            if args[1][0] == "-":
                index_cnt = 1

        # the first argument -- LFN.
        if 1 + index_cnt <= len(args) <= 2 + index_cnt:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1 + index_cnt) or (len(args) == 2 + index_cnt and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2 + index_cnt:
                    cur_path = args[1 + index_cnt]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_chown(self, args):
        """Change owner of the given path

        usage: chown [-R] <owner> <path>
        """

        argss = args.split()
        recursive = False
        if len(argss) == 0:
            print(self.do_chown.__doc__)
            return
        if argss[0] == "-R":
            recursive = True
            del argss[0]
        if len(argss) != 2:
            print(self.do_chown.__doc__)
            return
        owner = argss[0]
        path = argss[1]
        lfn = self.getPath(path)
        pathDict = {}
        pathDict[lfn] = owner

        try:
            result = self.fc.changePathOwner(pathDict, recursive)
            if not result["OK"]:
                print("Error:", result["Message"])
                return
            if lfn in result["Value"]["Failed"]:
                print("Error:", result["Value"]["Failed"][lfn])
                return
        except Exception as x:
            print("Exception:", str(x))

    def complete_chown(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        index_counter = 0 + 1

        if "-R" in args:
            index_counter = 1 + 1

        # the first argument -- LFN.
        if (1 + index_counter) <= len(args) <= (2 + index_counter):
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1 + index_counter) or (len(args) == 2 + index_counter and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2 + index_counter:
                    cur_path = args[1 + index_counter]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_chgrp(self, args):
        """Change group of the given path

        usage: chgrp [-R] <group> <path>
        """

        argss = args.split()
        recursive = False
        if len(argss) == 0:
            print(self.do_chgrp.__doc__)
            return
        if argss[0] == "-R":
            recursive = True
            del argss[0]
        if len(argss) != 2:
            print(self.do_chgrp.__doc__)
            return
        group = argss[0]
        path = argss[1]
        lfn = self.getPath(path)
        pathDict = {}
        pathDict[lfn] = group

        try:
            result = self.fc.changePathGroup(pathDict, recursive)
            if not result["OK"]:
                print("Error:", result["Message"])
                return
            if lfn in result["Value"]["Failed"]:
                print("Error:", result["Value"]["Failed"][lfn])
                return
        except Exception as x:
            print("Exception:", str(x))

    def complete_chgrp(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        index_counter = 0 + 1

        if "-R" in args:
            index_counter = 1 + 1

        # the first argument -- LFN.
        if (1 + index_counter) <= len(args) <= (2 + index_counter):
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1 + index_counter) or (len(args) == 2 + index_counter and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2 + index_counter:
                    cur_path = args[1 + index_counter]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_chmod(self, args):
        """Change permissions of the given path
        usage: chmod [-R] <mode> <path>
        """

        argss = args.split()
        recursive = False
        if len(argss) < 2:
            print(self.do_chmod.__doc__)
            return
        if argss[0] == "-R":
            recursive = True
            del argss[0]
        mode = argss[0]
        path = argss[1]
        lfn = self.getPath(path)
        pathDict = {}
        # treat mode as octal
        pathDict[lfn] = int(mode, base=8)

        try:
            result = self.fc.changePathMode(pathDict, recursive)
            if not result["OK"]:
                print("Error:", result["Message"])
                return
            if lfn in result["Value"]["Failed"]:
                print("Error:", result["Value"]["Failed"][lfn])
                return
        except Exception as x:
            print("Exception:", str(x))

    def complete_chmod(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        index_counter = 0 + 1

        if "-R" in args:
            index_counter = 1 + 1

        # the first argument -- LFN.
        if (1 + index_counter) <= len(args) <= (2 + index_counter):
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1 + index_counter) or (len(args) == 2 + index_counter and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2 + index_counter:
                    cur_path = args[1 + index_counter]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_size(self, args):
        """Get file or directory size. If -l switch is specified, get also the total
        size per Storage Element

        usage: size [-l] [-f] <lfn>|<dir_path>

        Switches:
           -l  long output including per SE report
           -f  use raw file information and not the storage tables
        """

        argss = args.split()
        _long = False
        fromFiles = False
        if len(argss) > 0:
            if argss[0] == "-l":
                _long = True
                del argss[0]
        if len(argss) > 0:
            if argss[0] == "-f":
                fromFiles = True
                del argss[0]

        if len(argss) == 1:
            path = argss[0]
            if path == ".":
                path = self.cwd
        else:
            path = self.cwd
        path = self.getPath(path)

        try:
            result = self.fc.isFile(path)
            if not result["OK"]:
                print("Error:", result["Message"])
            if result["Value"]["Successful"]:
                if result["Value"]["Successful"][path]:
                    print("lfn:", path)
                    result = self.fc.getFileSize(path)
                    if result["OK"]:
                        if result["Value"]["Successful"]:
                            print("Size:", result["Value"]["Successful"][path])
                        else:
                            print("File size failed:", result["Value"]["Failed"][path])
                    else:
                        print("File size failed:", result["Message"])
                else:
                    print("directory:", path)
                    result = self.fc.getDirectorySize(path, _long, fromFiles)
                    if result["OK"]:
                        if result["Value"]["Successful"]:
                            print(
                                "Logical Size:",
                                int_with_commas(result["Value"]["Successful"][path]["LogicalSize"]),
                                "Files:",
                                result["Value"]["Successful"][path]["LogicalFiles"],
                                "Directories:",
                                result["Value"]["Successful"][path]["LogicalDirectories"],
                            )
                            if _long:
                                fields = ["StorageElement", "Size", "Replicas"]
                                values = []
                                if "PhysicalSize" in result["Value"]["Successful"][path]:
                                    print()
                                    totalSize = result["Value"]["Successful"][path]["PhysicalSize"]["TotalSize"]
                                    totalFiles = result["Value"]["Successful"][path]["PhysicalSize"]["TotalFiles"]
                                    for se, sdata in result["Value"]["Successful"][path]["PhysicalSize"].items():
                                        if not se.startswith("Total"):
                                            size = sdata["Size"]
                                            nfiles = sdata["Files"]
                                            values.append((se, int_with_commas(size), str(nfiles)))
                                    values.append(("Total", int_with_commas(totalSize), str(totalFiles)))
                                    printTable(fields, values)
                            if "QueryTime" in result["Value"]:
                                print("Query time %.2f sec" % result["Value"]["QueryTime"])
                        else:
                            print("Directory size failed:", result["Value"]["Failed"][path])
                    else:
                        print("Directory size failed:", result["Message"])
            else:
                print("Failed to determine path type")
        except Exception as x:
            print("Size failed: ", x)

    def complete_size(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        index_counter = 0

        if "-l" in args:
            index_counter = 1

        # the first argument -- LFN.
        if (1 + index_counter) <= len(args) <= (2 + index_counter):
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1 + index_counter) or (len(args) == 2 + index_counter and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2 + index_counter:
                    cur_path = args[1 + index_counter]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def do_guid(self, args):
        """Get the file GUID

        usage: guid <lfn>
        """

        argss = args.split()
        if len(argss) == 0:
            print(self.do_guid.__doc__)
            return
        path = argss[0]
        path = self.getPath(path)
        try:
            result = self.fc.getFileMetadata(path)
            if result["OK"]:
                if result["Value"]["Successful"]:
                    print("GUID:", result["Value"]["Successful"][path]["GUID"])
                else:
                    print("ERROR: getting guid failed")
            else:
                print("ERROR:", result["Message"])
        except Exception as x:
            print("guid failed: ", x)

    def complete_guid(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    ##################################################################################
    #  Metadata methods

    def do_meta(self, args):
        """Metadata related operations

        Usage:
          meta index [-d|-f|-r] <metaname> [<metatype>]  - add new metadata index. Possible types are:
                                                           'int', 'float', 'string', 'date';
                                                         -d  directory metadata
                                                         -f  file metadata
                                                         -r  remove the specified metadata index
          meta set <path> <metaname> <metavalue> [<metaname> <metavalue> ...]- set metadata values for directory or file
          meta remove <path> <metaname> [<metaname> ...] - remove metadata values for directory or file
          meta get [-e] [<path>] - get metadata for the given directory or file
          meta tags <path> <metaname> where <meta_selection> - get values (tags) of the given metaname compatible with
                                                        the metadata selection
          meta show - show all defined metadata indice

        """
        argss = args.split()
        if len(argss) == 0:
            print(self.do_meta.__doc__)
            return
        option = argss[0]
        del argss[0]
        if option == "set":
            if len(argss) < 3 or len(argss) % 2 != 1:
                print(self.do_meta.__doc__)
                return
            return self.setMeta(argss)
        elif option == "get":
            return self.getMeta(argss)
        elif option[:3] == "tag":
            # TODO
            if len(argss) == 0:
                print(self.do_meta.__doc__)
                return
            return self.metaTag(argss)
        elif option == "index":
            if len(argss) < 1:
                print(self.do_meta.__doc__)
                return
            return self.registerMeta(argss)
        elif option == "metaset":
            # TODO
            if len(argss) == 0:
                print(self.do_meta.__doc__)
                return
            return self.registerMetaset(argss)
        elif option == "show":
            return self.showMeta()
        elif option == "remove" or option == "rm":
            if len(argss) < 2:
                print(self.do_meta.__doc__)
                return
            return self.removeMeta(argss)
        else:
            print("Unknown option:", option)

    # auto completion for ``meta``
    # TODO: what's the doc for metaset?
    _available_meta_cmd = ["set", "get", "tag", "tags", "index", "metaset", "show", "rm", "remove"]
    _meta_cmd_need_lfn = ["set", "get", "rm", "remove"]

    def complete_meta(self, text, line, begidx, endidx):
        result = []
        args = line.split()
        if len(args) >= 2 and (args[1] in self._available_meta_cmd):
            # if the sub command is not in self._meta_cmd_need_lfn
            # Don't need any auto completion
            if args[1] in self._meta_cmd_need_lfn:
                # TODO
                if len(args) == 2:
                    cur_path = ""
                elif len(args) > 2:
                    # If the line ends with ' '
                    # this means a new parameter begin.
                    if line.endswith(" "):
                        cur_path = ""
                    else:
                        cur_path = args[-1]

                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)
                pass
            return result

        result = [i for i in self._available_meta_cmd if i.startswith(text)]
        return result

    def removeMeta(self, argss):
        """Remove the specified metadata for a directory or file"""
        apath = argss[0]
        path = self.getPath(apath)
        if len(argss) < 2:
            print("Error: no metadata is specified for removal")
            return

        metadata = argss[1:]
        metaDict = {path: metadata}
        result = self.fc.removeMetadata(metaDict)
        if not result["OK"]:
            print("Error:", result["Message"])
            if "FailedMetadata" in result:
                for meta, error in result["FailedMetadata"]:
                    print(meta, ";", error)

    def setMeta(self, argss):
        """Set metadata value for a directory"""
        if len(argss) < 3 or len(argss) % 2 != 1:
            print(
                "Error: command requires at least 3 arguments (or odd number of arguments > 3), %d given" % len(argss)
            )
            return
        path = argss[0]
        if path == ".":
            path = self.cwd
        elif path[0] != "/":
            path = self.cwd + "/" + path
        path = self.getPath(path)
        del argss[0]
        meta = argss[::2]
        value = argss[1::2]
        metadict = {meta[n]: value[n] for n in range(len(meta))}
        print(path, metadict)
        result = self.fc.setMetadata(path, metadict)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))

    def getMeta(self, argss):
        """Get metadata for the given directory"""
        expandFlag = False
        dirFlag = True
        if len(argss) == 0:
            path = "."
        else:
            if argss[0] == "-e":
                expandFlag = True
                del argss[0]
            if len(argss) == 0:
                path = "."
            else:
                path = argss[0]
                dirFlag = False
        if path == ".":
            path = self.cwd
        elif path[0] != "/":
            path = self.getPath(path)

        path = path.rstrip("/")

        if not dirFlag:
            # Have to decide if it is a file or not
            result = self.fc.isFile(path)
            if not result["OK"]:
                print("ERROR: Failed to contact the catalog")
            if not result["Value"]["Successful"]:
                print("ERROR: Path not found")
            dirFlag = not result["Value"]["Successful"][path]

        if dirFlag:
            result = self.fc.getDirectoryUserMetadata(path)
            if not result["OK"]:
                print(("Error: %s" % result["Message"]))
                return
            if result["Value"]:
                metaDict = result["MetadataOwner"]
                metaTypeDict = result["MetadataType"]
                for meta, value in result["Value"].items():
                    setFlag = metaDict[meta] != "OwnParameter" and metaTypeDict[meta] == "MetaSet"
                    prefix = ""
                    if setFlag:
                        prefix = "+"
                    if metaDict[meta] == "ParentMetadata":
                        prefix += "*"
                        print((prefix + meta).rjust(20), ":", value)
                    elif metaDict[meta] == "OwnMetadata":
                        prefix += "!"
                        print((prefix + meta).rjust(20), ":", value)
                    else:
                        print(meta.rjust(20), ":", value)
                    if setFlag and expandFlag:
                        result = self.fc.getMetadataSet(value, expandFlag)
                        if not result["OK"]:
                            print(("Error: %s" % result["Message"]))
                            return
                        for m, v in result["Value"].items():
                            print(" " * 10, m.rjust(20), ":", v)
            else:
                print("No metadata defined for directory")
        else:
            result = self.fc.getFileUserMetadata(path)
            if not result["OK"]:
                print(("Error: %s" % result["Message"]))
                return
            if result["Value"]:
                for meta, value in result["Value"].items():
                    print(meta.rjust(20), ":", value)
            else:
                print("No metadata found")

    def metaTag(self, argss):
        """Get values of a given metadata tag compatible with the given selection"""
        path = argss[0]
        del argss[0]
        tag = argss[0]
        del argss[0]
        path = self.getPath(path)

        # Evaluate the selection dictionary
        metaDict = {}
        if argss:
            if argss[0].lower() == "where":
                result = self.fc.getMetadataFields()
                if not result["OK"]:
                    print(("Error: %s" % result["Message"]))
                    return
                if not result["Value"]:
                    print("Error: no metadata fields defined")
                    return
                typeDictfm = result["Value"]["FileMetaFields"]
                typeDict = result["Value"]["DirectoryMetaFields"]

                del argss[0]
                for arg in argss:
                    try:
                        name, value = arg.split("=")
                        if name not in typeDict:
                            if name not in typeDictfm:
                                print("Error: metadata field %s not defined" % name)
                            else:
                                print("No support for meta data at File level yet: %s" % name)
                            return
                        mtype = typeDict[name]
                        mvalue = value
                        if mtype[0:3].lower() == "int":
                            mvalue = int(value)
                        if mtype[0:5].lower() == "float":
                            mvalue = float(value)
                        metaDict[name] = mvalue
                    except Exception as x:
                        print("Error:", str(x))
                        return
            else:
                print("Error: WHERE keyword is not found after the metadata tag name")
                return

        result = self.fc.getCompatibleMetadata(metaDict, path)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
            return
        tagDict = result["Value"]
        if tag in tagDict:
            if tagDict[tag]:
                print("Possible values for %s:" % tag)
                for v in tagDict[tag]:
                    print(v)
            else:
                print("No compatible values found for %s" % tag)

    def showMeta(self):
        """Show defined metadata indices"""
        result = self.fc.getMetadataFields()
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
        else:
            if not result["Value"]:
                print("No entries found")
            else:
                for meta, _type in result["Value"].items():
                    print(meta.rjust(20), ":", _type)

    def registerMeta(self, argss):
        """Add metadata field."""

        if len(argss) < 2:
            print("Unsufficient number of arguments")
            return

        fdType = "-d"
        removeFlag = False
        if argss[0].lower() in ["-d", "-f"]:
            fdType = argss[0]
            del argss[0]
        if argss[0].lower() == "-r":
            removeFlag = True
            del argss[0]

        if len(argss) < 2 and not removeFlag:
            print("Unsufficient number of arguments")
            return

        mname = argss[0]
        if removeFlag:
            result = self.fc.deleteMetadataField(mname)
            if not result["OK"]:
                print("Error:", result["Message"])
            return

        mtype = argss[1]

        if mtype.lower()[:3] == "int":
            rtype = "INT"
        elif mtype.lower()[:7] == "varchar":
            rtype = mtype
        elif mtype.lower() == "string":
            rtype = "VARCHAR(128)"
        elif mtype.lower() == "float":
            rtype = "FLOAT"
        elif mtype.lower() == "date":
            rtype = "DATETIME"
        elif mtype.lower() == "metaset":
            rtype = "MetaSet"
        else:
            print("Error: illegal metadata type %s" % mtype)
            return

        result = self.fc.addMetadataField(mname, rtype, fdType)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
        else:
            print("Added metadata field %s of type %s" % (mname, mtype))

    def registerMetaset(self, argss):
        """Add metadata set"""

        setDict = {}
        setName = argss[0]
        del argss[0]
        for arg in argss:
            key, value = arg.split("=")
            setDict[key] = value

        result = self.fc.addMetadataSet(setName, setDict)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
        else:
            print("Added metadata set %s" % setName)

    def do_find(self, args):
        """Find all files satisfying the given metadata information

        usage: find [-q] [-D] <path> <meta_name>=<meta_value> [<meta_name>=<meta_value>]
        """

        argss = args.split()
        if len(argss) < 1:
            print(self.do_find.__doc__)
            return

        verbose = True
        if argss[0] == "-q":
            verbose = False
            del argss[0]

        dirsOnly = False
        if argss[0] == "-D":
            dirsOnly = True
            del argss[0]

        path = argss[0]
        path = self.getPath(path)
        del argss[0]

        if argss:
            if argss[0][0] == "{":
                metaDict = eval(argss[0])
            else:
                result = self.__createQuery(" ".join(argss))
                if not result["OK"]:
                    print("Illegal metaQuery:", " ".join(argss), result["Message"])
                    return
                metaDict = result["Value"]
        else:
            metaDict = {}
        if verbose:
            print("Query:", metaDict)

        result = self.fc.findFilesByMetadata(metaDict, path)
        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
            return

        if result["Value"]:

            lfnList = result["Value"]
            if dirsOnly:
                listToPrint = set(os.path.dirname(fullpath) for fullpath in lfnList)
            else:
                listToPrint = lfnList

            for dir_ in listToPrint:
                print(dir_)

        else:
            if verbose:
                print("No matching data found")

        if verbose and "QueryTime" in result:
            print("QueryTime %.2f sec" % result["QueryTime"])

    def complete_find(self, text, line, begidx, endidx):
        result = []
        args = line.split()

        # skip "-q" optional switch
        if len(args) >= 2 and args[1] == "-q":
            if len(args) > 2 or line.endswith(" "):
                del args[1]

        # the first argument -- LFN.
        if 1 <= len(args) <= 2:
            # If last char is ' ',
            # this can be a new parameter.
            if (len(args) == 1) or (len(args) == 2 and (not line.endswith(" "))):
                cur_path = ""
                if len(args) == 2:
                    cur_path = args[1]
                result = self.lfn_dc.parse_text_line(text, cur_path, self.cwd)

        return result

    def __createQuery(self, args):
        """Create the metadata query out of the command line arguments"""
        argss = args.split()
        result = self.fc.getMetadataFields()

        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
            return None
        if not result["Value"]:
            print("Error: no metadata fields defined")
            return None
        typeDict = result["Value"]["FileMetaFields"]
        typeDict.update(result["Value"]["DirectoryMetaFields"])

        # Special meta tags
        typeDict.update(FILE_STANDARD_METAKEYS)

        mq = MetaQuery(typeDict=typeDict)
        return mq.setMetaQuery(argss)

    def do_dataset(self, args):
        """A set of dataset manipulation commands

        Usage:

          dataset add <dataset_name> <meta_query>          - add a new dataset definition
          dataset annotate <dataset_name> <annotation>     - add annotation to a dataset
          dataset show [-l] [<dataset_name>]               - show existing datasets
          dataset status <dataset_name>                    - display the dataset status
          dataset files <dataset_name>                     - show dataset files
          dataset rm <dataset_name>                        - remove dataset
          dataset check <dataset_name>                     - check if the dataset parameters are still valid
          dataset update <dataset_name>                    - update the dataset parameters
          dataset freeze <dataset_name>                    - fix the current contents of the dataset
          dataset release <dataset_name>                   - release the dynamic dataset
        """
        argss = args.split()
        if len(argss) == 0:
            print(self.do_dataset.__doc__)
            return
        command = argss[0]
        del argss[0]
        if command == "add":
            self.dataset_add(argss)
        elif command == "annotate":
            self.dataset_annotate(argss)
        elif command == "show":
            self.dataset_show(argss)
        elif command == "files":
            self.dataset_files(argss)
        elif command == "rm":
            self.dataset_rm(argss)
        elif command == "check":
            self.dataset_check(argss)
        elif command == "update":
            self.dataset_update(argss)
        elif command == "freeze":
            self.dataset_freeze(argss)
        elif command == "release":
            self.dataset_release(argss)
        elif command == "status":
            self.dataset_status(argss)

    def dataset_add(self, argss):
        """Add a new dataset"""
        datasetName = argss[0]
        metaSelections = " ".join(argss[1:])
        result = self.__createQuery(metaSelections)
        if not result["OK"]:
            print("Illegal metaQuery:", metaSelections)
            return
        metaDict = result["Value"]
        datasetName = self.getPath(datasetName)

        result = returnSingleResult(self.fc.addDataset({datasetName: metaDict}))
        if not result["OK"]:
            print("ERROR: failed to add dataset:", result["Message"])
        else:
            print("Successfully added dataset", datasetName)

    def dataset_annotate(self, argss):
        """Add a new dataset"""
        datasetName = argss[0]
        annotation = " ".join(argss[1:])
        datasetName = self.getPath(datasetName)

        result = returnSingleResult(self.fc.addDatasetAnnotation({datasetName: annotation}))
        if not result["OK"]:
            print("ERROR: failed to add annotation:", result["Message"])
        else:
            print("Successfully added annotation to", datasetName)

    def dataset_status(self, argss):
        """Display the dataset status"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.getDatasetParameters(datasetName))
        if not result["OK"]:
            print("ERROR: failed to get status of dataset:", result["Message"])
        else:
            parDict = result["Value"]
            for par, value in parDict.items():
                print(par.rjust(20), ":", value)

    def dataset_rm(self, argss):
        """Remove the given dataset"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.removeDataset(datasetName))
        if not result["OK"]:
            print("ERROR: failed to remove dataset:", result["Message"])
        else:
            print("Successfully removed dataset", datasetName)

    def dataset_check(self, argss):
        """check if the dataset parameters are still valid"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.checkDataset(datasetName))
        if not result["OK"]:
            print("ERROR: failed to check dataset:", result["Message"])
        else:
            changeDict = result["Value"]
            if not changeDict:
                print("Dataset is not changed")
            else:
                print("Dataset changed:")
                for par in changeDict:
                    print("   ", par, ": ", changeDict[par][0], "->", changeDict[par][1])

    def dataset_update(self, argss):
        """Update the given dataset parameters"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.updateDataset(datasetName))
        if not result["OK"]:
            print("ERROR: failed to update dataset:", result["Message"])
        else:
            print("Successfully updated dataset", datasetName)

    def dataset_freeze(self, argss):
        """Freeze the given dataset"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.freezeDataset(datasetName))
        if not result["OK"]:
            print("ERROR: failed to freeze dataset:", result["Message"])
        else:
            print("Successfully frozen dataset", datasetName)

    def dataset_release(self, argss):
        """Release the given dataset"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.releaseDataset(datasetName))
        if not result["OK"]:
            print("ERROR: failed to release dataset:", result["Message"])
        else:
            print("Successfully released dataset", datasetName)

    def dataset_files(self, argss):
        """Get the given dataset files"""
        datasetName = argss[0]
        result = returnSingleResult(self.fc.getDatasetFiles(datasetName))
        if not result["OK"]:
            print("ERROR: failed to get files for dataset:", result["Message"])
        else:
            lfnList = result["Value"]
            for lfn in lfnList:
                print(lfn)

    def dataset_show(self, argss):
        """Show existing requested datasets"""
        long_ = False
        if "-l" in argss:
            long_ = True
            del argss[argss.index("-l")]
        datasetName = "*"
        if len(argss) > 0:
            datasetName = argss[0]

        result = returnSingleResult(self.fc.getDatasets(datasetName))
        if not result["OK"]:
            print("ERROR: failed to get datasets")
            return

        datasetDict = result["Value"]
        if not long_:
            for dName in datasetDict.keys():
                print(dName)
        else:
            fields = ["Key", "Value"]
            datasets = list(datasetDict)
            dsAnnotations = {}
            resultAnno = returnSingleResult(self.fc.getDatasetAnnotation(datasets))
            if resultAnno["OK"]:
                dsAnnotations = resultAnno["Value"]
            for dName in datasets:
                records = []
                print("\n" + dName + ":")
                print("=" * (len(dName) + 1))
                for key, value in datasetDict[dName].items():
                    records.append([key, str(value)])
                if dName in dsAnnotations:
                    records.append(["Annotation", dsAnnotations[dName]])
                printTable(fields, records)

    def do_stats(self, args):
        """Get the catalog statistics

        Usage:
          stats
        """

        try:
            result = self.fc.getCatalogCounters()
        except AttributeError as x:
            print("Error: no statistics available for this type of catalog:", str(x))
            return

        if not result["OK"]:
            print(("Error: %s" % result["Message"]))
            return
        fields = ["Counter", "Number"]
        records = []
        for key, value in result["Value"].items():
            records.append((key, str(value)))
        printTable(fields, records)

    def do_rebuild(self, args):
        """Rebuild auxiliary tables

        Usage:
           rebuild <option>
        """

        argss = args.split()
        _option = argss[0]
        start = time.time()
        result = self.fc.rebuildDirectoryUsage(timeout=300)
        if not result["OK"]:
            print("Error:", result["Message"])
            return

        total = time.time() - start
        print("Directory storage info rebuilt in %.2f sec", total)

    def do_repair(self, args):
        """Repair catalog inconsistencies

        Usage:
           repair
        """

        start = time.time()
        result = self.fc.repairCatalog()
        if not result["OK"]:
            print("Error:", result["Message"])
            return

        for repType, repResult in result["Value"].items():
            print("%s repair: %s" % (repType, repResult))

        total = time.time() - start
        print("Catalog repair operation done in %.2f sec" % total)


if __name__ == "__main__":

    if len(sys.argv) > 2:
        print(FileCatalogClientCLI.__doc__)
        sys.exit(2)

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    catalogs = None
    if len(sys.argv) == 2:
        catalogs = [sys.argv[1]]
    fc = FileCatalog(catalogs=catalogs)
    cli = FileCatalogClientCLI(fc)
    if catalogs:
        print("Starting %s file catalog client", catalogs[0])
    cli.cmdloop()



""" Class that encapsulate few commonalities of DIRAC DBs
    (MySQL-based, SQLAlchemy-based, ES-based)
"""

from DIRAC import gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection


class DIRACDB(object):
    """Extended in DB, SQLAlchemyDB, ElasticDB"""

    def __init__(self, *args, **kwargs):
        """c'tor

        :param self: self reference
        """
        self.log = gLogger.getSubLogger(self.fullname)  # pylint: disable=no-member
        super(DIRACDB, self).__init__(*args, **kwargs)

    def getCSOption(self, optionName, defaultValue=None):
        cs_path = getDatabaseSection(self.fullname)  # pylint: disable=no-member
        return gConfig.getValue("/%s/%s" % (cs_path, optionName), defaultValue)



#! /usr/bin/env python
########################################################################
# File :    dirac-stager-stage-files
# Author :  Daniela Remenska
########################################################################
"""
Submit staging requests for a particular Storage Element! Default DIRAC JobID will be =0.
(not visible in the Job monitoring list though)

Usage:
  dirac-stager-stage-files <LFN> <SE>

Arguments:
  <LFN>: LFN to Stage (or local file with list of LFNs)
  <SE>:  Name of Storage Element

Example:
  $ dirac-stager-stage-files.py filesToStage.txt GRIDKA-RDST
  Stage request submitted for LFNs:
  ['/lhcb/LHCb/Collision12/FULL.DST/00020846/0002/00020846_00023458_1.full.dst',
  '/lhcb/LHCb/Collision12/FULL.DST/00020846/0003/00020846_00032669_1.full.dst',
  '/lhcb/LHCb/Collision12/FULL.DST/00020846/0003/00020846_00032666_1.full.dst']
  SE= GRIDKA-RDST
  You can check their status and progress with dirac-stager-monitor-file <LFN> <SE>

  $ dirac-stager-stage-files.py /lhcb/LHCb/Collision12/FULL.DST/00020846/0003/00020846_00032641_1.full.dst GRIDKA-RDST
  Stage request submitted for LFNs:
  ['/lhcb/LHCb/Collision12/FULL.DST/00020846/0003/00020846_00032641_1.full.dst']
  SE= GRIDKA-RDST
  You can check their status and progress with dirac-stager-monitor-file <LFN> <SE>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    args = Script.getPositionalArgs()

    if len(args) < 2:
        Script.showHelp()

    seName = args[1]
    fileName = args[0]

    import os
    from DIRAC import exit as DIRACExit, gLogger
    from DIRAC.Interfaces.API.Dirac import Dirac
    from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient

    stageLfns = {}

    if os.path.exists(fileName):
        try:
            lfnFile = open(fileName)
            lfns = [k.strip() for k in lfnFile.readlines()]
            lfnFile.close()
        except Exception:
            gLogger.exception("Can not open file", fileName)
            DIRACExit(-1)
    else:
        lfns = args[: len(args) - 1]

    stageLfns[seName] = lfns
    stagerClient = StorageManagerClient()

    res = stagerClient.setRequest(
        stageLfns, "WorkloadManagement", "updateJobFromStager@WorkloadManagement/JobStateUpdate", 0
    )  # fake JobID = 0
    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRACExit(-1)
    else:
        gLogger.notice("Stage request submitted for LFNs:\n %s" % lfns)
        gLogger.notice("SE= %s" % seName)
        gLogger.notice("You can check their status and progress with dirac-stager-monitor-file <LFN> <SE>")

    DIRACExit()


if __name__ == "__main__":
    main()



import six
import time
from six import StringIO
import copy
import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.List import fromChar
from DIRAC.Core.Utilities.ModuleFactory import ModuleFactory
from DIRAC.Core.Utilities.DErrno import ETSDATA, ETSUKN
from DIRAC.Interfaces.API.Job import Job
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.TransformationSystem.Client.TaskManager import TaskBase
from DIRAC.WorkloadManagementSystem.Client.WMSClient import WMSClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient


class WorkflowTasks(TaskBase):
    """Handles jobs"""

    def __init__(
        self,
        transClient=None,
        logger=None,
        submissionClient=None,
        jobMonitoringClient=None,
        outputDataModule=None,
        jobClass=None,
        opsH=None,
        destinationPlugin=None,
        ownerDN=None,
        ownerGroup=None,
    ):
        """Generates some default objects.
        jobClass is by default "DIRAC.Interfaces.API.Job.Job". An extension of it also works:
        VOs can pass in their job class extension, if present
        """

        if not logger:
            logger = gLogger.getSubLogger("WorkflowTasks")

        super(WorkflowTasks, self).__init__(transClient, logger)

        useCertificates = True if (bool(ownerDN) and bool(ownerGroup)) else False
        if not submissionClient:
            self.submissionClient = WMSClient(
                useCertificates=useCertificates, delegatedDN=ownerDN, delegatedGroup=ownerGroup
            )
        else:
            self.submissionClient = submissionClient

        if not jobMonitoringClient:
            self.jobMonitoringClient = JobMonitoringClient()
        else:
            self.jobMonitoringClient = jobMonitoringClient

        if not jobClass:
            self.jobClass = Job
        else:
            self.jobClass = jobClass

        if not opsH:
            self.opsH = Operations()
        else:
            self.opsH = opsH

        if not outputDataModule:
            self.outputDataModule = self.opsH.getValue("Transformations/OutputDataModule", "")
        else:
            self.outputDataModule = outputDataModule

        if not destinationPlugin:
            self.destinationPlugin = self.opsH.getValue("Transformations/DestinationPlugin", "BySE")
        else:
            self.destinationPlugin = destinationPlugin

        self.destinationPlugin_o = None

        self.outputDataModule_o = None

    def prepareTransformationTasks(
        self, transBody, taskDict, owner="", ownerGroup="", ownerDN="", bulkSubmissionFlag=False
    ):
        """Prepare tasks, given a taskDict, that is created (with some manipulation) by the DB
            jobClass is by default "DIRAC.Interfaces.API.Job.Job". An extension of it also works.


        :param str transBody: transformation job template
        :param dict taskDict: dictionary of per task parameters
        :param str owner: owner of the transformation
        :param str ownerGroup: group of the owner of the transformation
        :param str ownerDN: DN of the owner of the transformation
        :param bool bulkSubmissionFlag: flag for using bulk submission or not

        :return: S_OK/S_ERROR with updated taskDict
        """

        if (not owner) or (not ownerGroup):
            res = getProxyInfo(False, False)
            if not res["OK"]:
                return res
            proxyInfo = res["Value"]
            owner = proxyInfo["username"]
            ownerGroup = proxyInfo["group"]

        if not ownerDN:
            res = getDNForUsername(owner)
            if not res["OK"]:
                return res
            ownerDN = res["Value"][0]

        if bulkSubmissionFlag:
            return self.__prepareTasksBulk(transBody, taskDict, owner, ownerGroup, ownerDN)
        # not a bulk submission
        return self.__prepareTasks(transBody, taskDict, owner, ownerGroup, ownerDN)

    def __prepareTasksBulk(self, transBody, taskDict, owner, ownerGroup, ownerDN):
        """Prepare transformation tasks with a single job object for bulk submission

        :param str transBody: transformation job template
        :param dict taskDict: dictionary of per task parameters
        :param str owner: owner of the transformation
        :param str ownerGroup: group of the owner of the transformation
        :param str ownerDN: DN of the owner of the transformation

        :return: S_OK/S_ERROR with updated taskDict
        """
        if taskDict:
            transID = list(taskDict.values())[0]["TransformationID"]
        else:
            return S_OK({})

        method = "__prepareTasksBulk"
        startTime = time.time()

        # Prepare the bulk Job object with common parameters
        oJob = self.jobClass(transBody)
        self._logVerbose("Setting job owner:group to %s:%s" % (owner, ownerGroup), transID=transID, method=method)
        oJob.setOwner(owner)
        oJob.setOwnerGroup(ownerGroup)
        oJob.setOwnerDN(ownerDN)

        try:
            site = oJob.workflow.findParameter("Site").getValue()
        except AttributeError:
            site = None
        jobType = oJob.workflow.findParameter("JobType").getValue()
        transGroup = str(transID).zfill(8)

        # Verify that the JOB_ID parameter is added to the workflow
        if not oJob.workflow.findParameter("JOB_ID"):
            oJob._addParameter(oJob.workflow, "JOB_ID", "string", "00000000", "Initial JOB_ID")

        if oJob.workflow.findParameter("PRODUCTION_ID"):
            oJob._setParamValue("PRODUCTION_ID", str(transID).zfill(8))  # pylint: disable=protected-access
        else:
            oJob._addParameter(
                oJob.workflow,  # pylint: disable=protected-access
                "PRODUCTION_ID",
                "string",
                str(transID).zfill(8),
                "Production ID",
            )
        oJob.setType(jobType)
        self._logVerbose("Adding default transformation group of %s" % (transGroup), transID=transID, method=method)
        oJob.setJobGroup(transGroup)

        clinicPath = self._checkSickTransformations(transID)
        if clinicPath:
            self._handleHospital(oJob, clinicPath)

        # Collect per job parameters sequences
        paramSeqDict = {}
        # tasks must be sorted because we use bulk submission and we must find the correspondance
        for taskID in sorted(taskDict):
            paramsDict = taskDict[taskID]
            seqDict = {}

            if site is not None:
                paramsDict["Site"] = site
            paramsDict["JobType"] = jobType

            # Handle destination site
            sites = self._handleDestination(paramsDict)
            if not sites:
                self._logError("Could not get a list a sites", transID=transID, method=method)
                return S_ERROR(ETSUKN, "Can not evaluate destination site")
            else:
                self._logVerbose("Setting Site: ", str(sites), transID=transID, method=method)
                seqDict["Site"] = sites

            seqDict["JobName"] = self._transTaskName(transID, taskID)
            seqDict["JOB_ID"] = str(taskID).zfill(8)

            self._logDebug(
                "TransID: %s, TaskID: %s, paramsDict: %s" % (transID, taskID, str(paramsDict)),
                transID=transID,
                method=method,
            )

            # Handle Input Data
            inputData = paramsDict.get("InputData")
            if inputData:
                if isinstance(inputData, six.string_types):
                    inputData = inputData.replace(" ", "").split(";")
                self._logVerbose("Setting input data to %s" % inputData, transID=transID, method=method)
                seqDict["InputData"] = inputData
            elif paramSeqDict.get("InputData") is not None:
                self._logError("Invalid mixture of jobs with and without input data")
                return S_ERROR(ETSDATA, "Invalid mixture of jobs with and without input data")

            for paramName, paramValue in paramsDict.items():
                if paramName not in ("InputData", "Site", "TargetSE"):
                    if paramValue:
                        self._logVerbose("Setting %s to %s" % (paramName, paramValue), transID=transID, method=method)
                        seqDict[paramName] = paramValue

            outputParameterList = []
            if self.outputDataModule:
                res = self.getOutputData(
                    {
                        "Job": oJob._toXML(),  # pylint: disable=protected-access
                        "TransformationID": transID,
                        "TaskID": taskID,
                        "InputData": inputData,
                    }
                )
                if not res["OK"]:
                    self._logError("Failed to generate output data", res["Message"], transID=transID, method=method)
                    continue
                for name, output in res["Value"].items():
                    seqDict[name] = output
                    outputParameterList.append(name)
                    if oJob.workflow.findParameter(name):
                        oJob._setParamValue(name, "%%(%s)s" % name)  # pylint: disable=protected-access
                    else:
                        oJob._addParameter(
                            oJob.workflow, name, "JDL", "%%(%s)s" % name, name  # pylint: disable=protected-access
                        )

            for pName, seq in seqDict.items():
                paramSeqDict.setdefault(pName, []).append(seq)

        for paramName, paramSeq in paramSeqDict.items():
            if paramName in ["JOB_ID", "PRODUCTION_ID", "InputData"] + outputParameterList:
                res = oJob.setParameterSequence(paramName, paramSeq, addToWorkflow=paramName)
            else:
                res = oJob.setParameterSequence(paramName, paramSeq)
            if not res["OK"]:
                return res

        if taskDict:
            self._logInfo("Prepared %d tasks" % len(taskDict), transID=transID, method=method, reftime=startTime)

        taskDict["BulkJobObject"] = oJob
        return S_OK(taskDict)

    def __prepareTasks(self, transBody, taskDict, owner, ownerGroup, ownerDN):
        """Prepare transformation tasks with a job object per task

        :param str transBody: transformation job template
        :param dict taskDict: dictionary of per task parameters
        :param owner: owner of the transformation
        :param str ownerGroup: group of the owner of the transformation
        :param str ownerDN: DN of the owner of the transformation

        :return:  S_OK/S_ERROR with updated taskDict
        """
        if taskDict:
            transID = list(taskDict.values())[0]["TransformationID"]
        else:
            return S_OK({})

        method = "__prepareTasks"
        startTime = time.time()

        oJobTemplate = self.jobClass(transBody)
        oJobTemplate.setOwner(owner)
        oJobTemplate.setOwnerGroup(ownerGroup)
        oJobTemplate.setOwnerDN(ownerDN)

        try:
            site = oJobTemplate.workflow.findParameter("Site").getValue()
        except AttributeError:
            site = None
        jobType = oJobTemplate.workflow.findParameter("JobType").getValue()
        templateOK = False
        getOutputDataTiming = 0.0

        for taskID, paramsDict in taskDict.items():
            # Create a job for each task and add it to the taskDict
            if not templateOK:
                templateOK = True
                # Update the template with common information
                self._logVerbose("Job owner:group to %s:%s" % (owner, ownerGroup), transID=transID, method=method)
                transGroup = str(transID).zfill(8)
                self._logVerbose(
                    "Adding default transformation group of %s" % (transGroup), transID=transID, method=method
                )
                oJobTemplate.setJobGroup(transGroup)
                if oJobTemplate.workflow.findParameter("PRODUCTION_ID"):
                    oJobTemplate._setParamValue("PRODUCTION_ID", str(transID).zfill(8))
                else:
                    oJobTemplate._addParameter(
                        oJobTemplate.workflow, "PRODUCTION_ID", "string", str(transID).zfill(8), "Production ID"
                    )
                if not oJobTemplate.workflow.findParameter("JOB_ID"):
                    oJobTemplate._addParameter(oJobTemplate.workflow, "JOB_ID", "string", "00000000", "Initial JOB_ID")

            if site is not None:
                paramsDict["Site"] = site
            paramsDict["JobType"] = jobType
            # Now create the job from the template
            oJob = copy.deepcopy(oJobTemplate)
            constructedName = self._transTaskName(transID, taskID)
            self._logVerbose("Setting task name to %s" % constructedName, transID=transID, method=method)
            oJob.setName(constructedName)
            oJob._setParamValue("JOB_ID", str(taskID).zfill(8))
            inputData = None

            self._logDebug(
                "TransID: %s, TaskID: %s, paramsDict: %s" % (transID, taskID, str(paramsDict)),
                transID=transID,
                method=method,
            )

            # These helper functions do the real job
            sites = self._handleDestination(paramsDict)
            if not sites:
                self._logError("Could not get a list a sites", transID=transID, method=method)
                paramsDict["TaskObject"] = ""
                continue
            else:
                self._logDebug("Setting Site: ", str(sites), transID=transID, method=method)
                res = oJob.setDestination(sites)
                if not res["OK"]:
                    self._logError("Could not set the site: %s" % res["Message"], transID=transID, method=method)
                    paramsDict["TaskObject"] = ""
                    continue

            self._handleInputs(oJob, paramsDict)
            self._handleRest(oJob, paramsDict)

            clinicPath = self._checkSickTransformations(transID)
            if clinicPath:
                self._handleHospital(oJob, clinicPath)

            paramsDict["TaskObject"] = ""
            if self.outputDataModule:
                getOutputDataTiming -= time.time()
                res = self.getOutputData(
                    {"Job": oJob._toXML(), "TransformationID": transID, "TaskID": taskID, "InputData": inputData}
                )
                getOutputDataTiming += time.time()
                if not res["OK"]:
                    self._logError("Failed to generate output data", res["Message"], transID=transID, method=method)
                    continue
                for name, output in res["Value"].items():
                    oJob._addJDLParameter(name, ";".join(output))
            paramsDict["TaskObject"] = oJob
        if taskDict:
            self._logVerbose(
                "Average getOutputData time: %.1f per task" % (getOutputDataTiming / len(taskDict)),
                transID=transID,
                method=method,
            )
            self._logInfo("Prepared %d tasks" % len(taskDict), transID=transID, method=method, reftime=startTime)
        return S_OK(taskDict)

    #############################################################################

    def _handleDestination(self, paramsDict):
        """Handle Sites and TargetSE in the parameters"""

        try:
            sites = ["ANY"]
            if paramsDict["Site"]:
                # 'Site' comes from the XML and therefore is ; separated
                sites = fromChar(paramsDict["Site"], sepChar=";")
        except KeyError:
            pass

        if self.destinationPlugin_o:
            destinationPlugin_o = self.destinationPlugin_o
        else:
            res = self.__generatePluginObject(self.destinationPlugin)
            if not res["OK"]:
                self._logFatal("Could not generate a destination plugin object")
                return res
            destinationPlugin_o = res["Value"]
            self.destinationPlugin_o = destinationPlugin_o

        destinationPlugin_o.setParameters(paramsDict)
        destSites = destinationPlugin_o.run()
        if not destSites:
            return sites

        # Now we need to make the AND with the sites, if defined
        if sites != ["ANY"]:
            # Need to get the AND
            destSites &= set(sites)

        return list(destSites)

    def _handleInputs(self, oJob, paramsDict):
        """set job inputs (+ metadata)"""
        inputData = paramsDict.get("InputData")
        transID = paramsDict["TransformationID"]
        if inputData:
            self._logVerbose("Setting input data to %s" % inputData, transID=transID, method="_handleInputs")
            res = oJob.setInputData(inputData)
            if not res["OK"]:
                self._logError("Could not set the inputs: %s" % res["Message"], transID=transID, method="_handleInputs")

    def _handleRest(self, oJob, paramsDict):
        """add as JDL parameters all the other parameters that are not for inputs or destination"""
        transID = paramsDict["TransformationID"]
        for paramName, paramValue in paramsDict.items():
            if paramName not in ("InputData", "Site", "TargetSE"):
                if paramValue:
                    self._logDebug("Setting %s to %s" % (paramName, paramValue), transID=transID, method="_handleRest")
                    oJob._addJDLParameter(paramName, paramValue)

    def _checkSickTransformations(self, transID):
        """Check if the transformation is in the transformations to be processed at Hospital or Clinic"""
        transID = int(transID)
        clinicPath = "Hospital"
        if transID in set(int(x) for x in self.opsH.getValue(os.path.join(clinicPath, "Transformations"), [])):
            return clinicPath
        if "Clinics" in self.opsH.getSections("Hospital").get("Value", []):
            basePath = os.path.join("Hospital", "Clinics")
            clinics = self.opsH.getSections(basePath)["Value"]
            for clinic in clinics:
                clinicPath = os.path.join(basePath, clinic)
                if transID in set(int(x) for x in self.opsH.getValue(os.path.join(clinicPath, "Transformations"), [])):
                    return clinicPath
        return None

    def _handleHospital(self, oJob, clinicPath):
        """Optional handle of hospital/clinic jobs"""
        if not clinicPath:
            return
        oJob.setInputDataPolicy("download", dataScheduling=False)

        # Check first for a clinic, if not it must be the general hospital
        hospitalSite = self.opsH.getValue(os.path.join(clinicPath, "ClinicSite"), "")
        hospitalCEs = self.opsH.getValue(os.path.join(clinicPath, "ClinicCE"), [])
        # If not found, get the hospital parameters
        if not hospitalSite:
            hospitalSite = self.opsH.getValue("Hospital/HospitalSite", "DIRAC.JobDebugger.ch")
        if not hospitalCEs:
            hospitalCEs = self.opsH.getValue("Hospital/HospitalCEs", [])

        oJob.setDestination(hospitalSite)
        if hospitalCEs:
            oJob._addJDLParameter("GridCE", hospitalCEs)

    def __generatePluginObject(self, plugin):
        """This simply instantiates the TaskManagerPlugin class with the relevant plugin name"""
        method = "__generatePluginObject"
        try:
            plugModule = __import__(self.pluginLocation, globals(), locals(), ["TaskManagerPlugin"])
        except ImportError as e:
            self._logException("Failed to import 'TaskManagerPlugin' %s: %s" % (plugin, e), method=method)
            return S_ERROR()
        try:
            plugin_o = getattr(plugModule, "TaskManagerPlugin")("%s" % plugin, operationsHelper=self.opsH)
            return S_OK(plugin_o)
        except AttributeError as e:
            self._logException("Failed to create %s(): %s." % (plugin, e), method=method)
            return S_ERROR()

    #############################################################################

    def getOutputData(self, paramDict):
        """Get the list of job output LFNs from the provided plugin"""
        if not self.outputDataModule_o:
            # Create the module object
            moduleFactory = ModuleFactory()

            moduleInstance = moduleFactory.getModule(self.outputDataModule, None)
            if not moduleInstance["OK"]:
                return moduleInstance
            self.outputDataModule_o = moduleInstance["Value"]
        # This is the "argument" to the module, set it and then execute
        self.outputDataModule_o.paramDict = paramDict
        return self.outputDataModule_o.execute()

    def submitTransformationTasks(self, taskDict):
        """Submit the tasks"""
        if "BulkJobObject" in taskDict:
            return self.__submitTransformationTasksBulk(taskDict)
        return self.__submitTransformationTasks(taskDict)

    def __submitTransformationTasksBulk(self, taskDict):
        """Submit jobs in one go with one parametric job"""
        if not taskDict:
            return S_OK(taskDict)
        startTime = time.time()

        method = "__submitTransformationTasksBulk"

        oJob = taskDict.pop("BulkJobObject")
        # we can only do this, once the job has been popped, or we _might_ crash
        transID = list(taskDict.values())[0]["TransformationID"]
        if oJob is None:
            self._logError("no bulk Job object found", transID=transID, method=method)
            return S_ERROR(ETSUKN, "No bulk job object provided for submission")

        result = self.submitTaskToExternal(oJob)
        if not result["OK"]:
            self._logError("Failed to submit tasks to external", transID=transID, method=method)
            return result

        jobIDList = result["Value"]
        if len(jobIDList) != len(taskDict):
            for task in taskDict.values():
                task["Success"] = False
            return S_ERROR(ETSUKN, "Submitted less number of jobs than requested tasks")
        # Get back correspondence with tasks sorted by ID
        for jobID, taskID in zip(jobIDList, sorted(taskDict)):
            taskDict[taskID]["ExternalID"] = jobID
            taskDict[taskID]["Success"] = True

        submitted = len(jobIDList)
        self._logInfo(
            "Submitted %d tasks to WMS in %.1f seconds" % (submitted, time.time() - startTime),
            transID=transID,
            method=method,
        )
        return S_OK(taskDict)

    def __submitTransformationTasks(self, taskDict):
        """Submit jobs one by one"""
        method = "__submitTransformationTasks"
        submitted = 0
        failed = 0
        startTime = time.time()
        for task in taskDict.values():
            transID = task["TransformationID"]
            if not task["TaskObject"]:
                task["Success"] = False
                failed += 1
                continue
            res = self.submitTaskToExternal(task["TaskObject"])
            if res["OK"]:
                task["ExternalID"] = res["Value"]
                task["Success"] = True
                submitted += 1
            else:
                self._logError("Failed to submit task to WMS", res["Message"], transID=transID, method=method)
                task["Success"] = False
                failed += 1
        if submitted:
            self._logInfo(
                "Submitted %d tasks to WMS in %.1f seconds" % (submitted, time.time() - startTime),
                transID=transID,
                method=method,
            )
        if failed:
            self._logError("Failed to submit %d tasks to WMS." % (failed), transID=transID, method=method)
        return S_OK(taskDict)

    def submitTaskToExternal(self, job):
        """Submits a single job (which can be a bulk one) to the WMS."""
        if isinstance(job, six.string_types):
            try:
                oJob = self.jobClass(job)
            except Exception as x:  # pylint: disable=broad-except
                self._logException("Failed to create job object", "", x)
                return S_ERROR("Failed to create job object")
        elif isinstance(job, self.jobClass):
            oJob = job
        else:
            self._logError("No valid job description found")
            return S_ERROR("No valid job description found")

        workflowFileObject = StringIO(oJob._toXML())
        jdl = oJob._toJDL(jobDescriptionObject=workflowFileObject)
        return self.submissionClient.submitJob(jdl, workflowFileObject)

    def updateTransformationReservedTasks(self, taskDicts):
        transID = None
        jobNames = [self._transTaskName(taskDict["TransformationID"], taskDict["TaskID"]) for taskDict in taskDicts]
        res = self.jobMonitoringClient.getJobs({"JobName": jobNames})
        if not res["OK"]:
            self._logError(
                "Failed to get task from WMS",
                res["Message"],
                transID=transID,
                method="updateTransformationReservedTasks",
            )
            return res
        jobNameIDs = {}
        for wmsID in res["Value"]:
            res = self.jobMonitoringClient.getJobSummary(int(wmsID))
            if not res["OK"]:
                self._logWarn(
                    "Failed to get task summary from WMS",
                    res["Message"],
                    transID=transID,
                    method="updateTransformationReservedTasks",
                )
            else:
                jobNameIDs[res["Value"]["JobName"]] = int(wmsID)

        noTask = list(set(jobNames) - set(jobNameIDs))
        return S_OK({"NoTasks": noTask, "TaskNameIDs": jobNameIDs})

    def getSubmittedTaskStatus(self, taskDicts):
        """
        Check the status of a list of tasks and return lists of taskIDs for each new status
        """
        method = "getSubmittedTaskStatus"

        if taskDicts:
            wmsIDs = [int(taskDict["ExternalID"]) for taskDict in taskDicts if int(taskDict["ExternalID"])]
            transID = taskDicts[0]["TransformationID"]
        else:
            return S_OK({})
        res = self.jobMonitoringClient.getJobsStatus(wmsIDs)
        if not res["OK"]:
            self._logWarn("Failed to get job status from the WMS system", transID=transID, method=method)
            return res
        statusDict = res["Value"]
        updateDict = {}
        for taskDict in taskDicts:
            taskID = taskDict["TaskID"]
            wmsID = int(taskDict["ExternalID"])
            if not wmsID:
                continue
            oldStatus = taskDict["ExternalStatus"]
            newStatus = statusDict.get(wmsID, {}).get("Status", "Removed")
            if oldStatus != newStatus:
                if newStatus == "Removed":
                    self._logVerbose(
                        "Production/Job %d/%d removed from WMS while it is in %s status" % (transID, taskID, oldStatus),
                        transID=transID,
                        method=method,
                    )
                    newStatus = "Failed"
                self._logVerbose(
                    "Setting job status for Production/Job %d/%d to %s" % (transID, taskID, newStatus),
                    transID=transID,
                    method=method,
                )
                updateDict.setdefault(newStatus, []).append(taskID)
        return S_OK(updateDict)

    def getSubmittedFileStatus(self, fileDicts):
        """
        Check the status of a list of files and return the new status of each LFN
        """
        if not fileDicts:
            return S_OK({})

        method = "getSubmittedFileStatus"

        # All files are from the same transformation
        transID = fileDicts[0]["TransformationID"]
        taskFiles = {}
        for fileDict in fileDicts:
            jobName = self._transTaskName(transID, fileDict["TaskID"])
            taskFiles.setdefault(jobName, {})[fileDict["LFN"]] = fileDict["Status"]

        res = self.updateTransformationReservedTasks(fileDicts)
        if not res["OK"]:
            self._logWarn("Failed to obtain taskIDs for files", transID=transID, method=method)
            return res
        noTasks = res["Value"]["NoTasks"]
        taskNameIDs = res["Value"]["TaskNameIDs"]

        updateDict = {}
        for jobName in noTasks:
            for lfn, oldStatus in taskFiles[jobName].items():
                if oldStatus != TransformationFilesStatus.UNUSED:
                    updateDict[lfn] = TransformationFilesStatus.UNUSED

        res = self.jobMonitoringClient.getJobsStatus(list(taskNameIDs.values()))
        if not res["OK"]:
            self._logWarn("Failed to get job status from the WMS system", transID=transID, method=method)
            return res
        statusDict = res["Value"]
        for jobName, wmsID in taskNameIDs.items():
            jobStatus = statusDict.get(wmsID, {}).get("Status")
            newFileStatus = {
                "Done": TransformationFilesStatus.PROCESSED,
                "Completed": TransformationFilesStatus.PROCESSED,
                "Failed": TransformationFilesStatus.UNUSED,
            }.get(jobStatus)
            if newFileStatus:
                for lfn, oldStatus in taskFiles[jobName].items():
                    if newFileStatus != oldStatus:
                        updateDict[lfn] = newFileStatus
        return S_OK(updateDict)



"""
    This is a comment
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

# $Source: /tmp/libdirac/tmp.stZoy15380/dirac/DIRAC3/DIRAC/Core/Workflow/test/WFSamples.py,v $

from DIRAC.Core.Workflow.Parameter import *
from DIRAC.Core.Workflow.Module import *
from DIRAC.Core.Workflow.Step import *
from DIRAC.Core.Workflow.Workflow import *
from DIRAC.Core.Workflow.WorkflowReader import *

""" Collection of objects for the testing"""

body1 = """class PrintOutput(object):
  # Warning!!
  # class name MUST me the same as a Type field
  # and MUST have method execute
  class_var = 0 # static class parameter

  def __init__(self):
    # constructor code
    self.enable = 1 #local class parameter
    self.version = 0
    self.debug = False
    self.message = 'empty message'

  def execute(self):
    # main execution function
    if self.enable :
      if self.debug:
        print 'Executing Module = ',str(type(self))
      print str(type(self.message)), self.message
    else:
      print 'Type1 - pass'


  def __del__(self):
    pass
"""

body2 = """class Summ(object):
  # Warning!!
  # class name MUST me the same as a Type field
  # and MUST have method execute
  class_var = 0 # static class parameter

  def __init__(self):
    # constructor code
    self.enable = 1 #local class parameter
    self.version = 0
    self.debug = False
    self.input1 = 0
    self.input2 = 0
    self.result = 0

  def execute(self):
    # main execution function
    if self.enable :
      if self.debug:
        print 'Executing Module = ',str(type(self))
      self.result=self.input1+self.input2
      if self.debug:
        print 'inputs are', self.input1, self.input2
        print 'Result is',self.result
    else:
      print str(type(self)), 'pass'


  def __del__(self):
    pass
"""

body3 = "from calendar import Calendar\n"


op1 = Parameter("enable", "True", "bool", "", "", True, False, "if False execution disabled")
op2 = Parameter("version", "1.25", "float", "", "", False, True, "we can get version of the module")
# op3 = Parameter("message","\'this is ugly module\'","string","","",False,False,"message for the printing")
op3 = Parameter("message", "@{inparam4}", "string", "", "", False, False, "message for the printing")
op4 = Parameter("debug", "False", "bool", "", "", True, False, "allows to print additional information")
op5 = Parameter("input1", "2", "int", "", "", True, False, "argument for addition")
op6 = Parameter("input2", "5", "int", "", "", True, False, "argument for addition")
op7 = Parameter("result", "0", "int", "", "", False, True, "argument for addition")

md1 = ModuleDefinition("PrintOutput")
md1.addParameter(op1)
md1.addParameter(op2)
md1.addParameter(op3)
md1.addParameter(op4)
md1.setDescription("Module to print imput messsage")
md1.setBody(body1)

md2 = ModuleDefinition("Summ")
md2.setBody(body2)
md2.addParameter(op1)
md2.addParameter(op2)
md2.addParameter(op4)
md2.addParameter(op5)
md2.addParameter(op6)
md2.addParameter(op7)

md3 = ModuleDefinition("PrintOutput")
md3.addParameter(op1)
md3.addParameter(op2)
md3.addParameter(op3)
md3.addParameter(op4)
md3.setDescription("Module to print imput messsage")
md3.setBody(body1)

sd1 = StepDefinition("TotalSumm")
sd1.addModule(md3)
sd1.addModule(md1)
sd1.addModule(md2)

sd1.addParameter(Parameter("enable_inst1", "True", "bool", "", "", True, True, "enabling instance 1"))
sd1.addParameter(Parameter("enable_inst2", "True", "bool", "", "", True, True, "enabling instance 2"))
sd1.addParameter(Parameter("enable_inst3", "True", "bool", "", "", True, True, "enabling instance 3"))
sd1.addParameter(Parameter("debug", "True", "bool", "", "", True, True, "enabling additional printing"))
sd1.addParameter(Parameter("input1", "3.8", "float", "", "", True, False, "input slot"))
sd1.addParameter(Parameter("input2", "8.2", "float", "", "", True, False, "input slot"))
sd1.addParameter(Parameter("input3", "2.0", "float", "", "", True, False, "input slot"))
sd1.addParameter(Parameter("result", "0.0", "float", "", "", False, True, "output"))
# sd1.append(Parameter("message","empty message","string","","",True, False, "output"))
sd1.addParameter(Parameter("message", "@{inparam4}", "string", "", "", True, False, "output"))

mi1 = sd1.createModuleInstance("Summ", "mi1")
mi2 = sd1.createModuleInstance("PrintOutput", "mi2")
mi3 = sd1.createModuleInstance("Summ", "mi3")
mi4 = sd1.createModuleInstance("PrintOutput", "mi4")
mi5 = sd1.createModuleInstance("Summ", "mi5")
mi6 = sd1.createModuleInstance("PrintOutput", "mi6")
mi7 = sd1.createModuleInstance("PrintOutput", "mi7")
mi8 = sd1.createModuleInstance("PrintOutput", "mi8")
mi9 = sd1.createModuleInstance("PrintOutput", "mi9")
mi10 = sd1.createModuleInstance("PrintOutput", "mi10")
mi11 = sd1.createModuleInstance("PrintOutput", "mi11")
mi12 = sd1.createModuleInstance("PrintOutput", "mi12")
mi13 = sd1.createModuleInstance("PrintOutput", "mi13")

mi1.findParameter("enable").link("self", "enable_inst1")
mi1.findParameter("debug").link("self", "debug")
mi1.findParameter("input1").link("self", "input1")
mi1.findParameter("input2").link("self", "input2")

mi2.findParameter("enable").link("mi1", "enable")
mi2.findParameter("debug").link("self", "debug")
# mi2.findParameter('message').link('mi1','result') # taken from the level of step

mi3.findParameter("enable").link("self", "enable_inst2")
mi3.findParameter("debug").link("self", "debug")
mi3.findParameter("input1").link("self", "input2")
mi3.findParameter("input2").link("self", "input3")

mi4.findParameter("enable").link("mi3", "enable")
mi4.findParameter("debug").link("self", "debug")
# mi4.findParameter('message').link('mi3','result') # taken from the previouse instance
mi4.findParameter("message").link("mi3", "result")  # taken from the previouse instance

mi5.findParameter("enable").link("self", "enable_inst3")
mi5.findParameter("debug").link("self", "debug")
mi5.findParameter("input1").link("mi1", "result")
mi5.findParameter("input2").link("mi3", "result")

mi6.findParameter("enable").link("mi5", "enable")
mi6.findParameter("debug").link("self", "debug")
# mi6.findParameter('message').link('mi5','result') # taken from the previouse instance (chain propagation)

mdouble = 2.3567
mi6.findParameter("message").setValue(mdouble, "float")
mlist = ["file1", "file2", "file3"]
mi7.findParameter("message").setValue(mlist, "list")
mdict = {"jack": 4098, "sape": 4139}
mi8.findParameter("message").setValue(mdict, "dict")
mtuple = (1, 2, 3, 4, 5)
mi9.findParameter("message").setValue(mtuple, "tuple")
mstring = (
    """\"Clever string of mine; WR.Output = \"Collection=\'EVTTAGS/TagCreator/1\' ADDRESS=\'/Event\' DATAFILE=\""""
)
# mstring = "Clever string of mine;"
mi10.findParameter("message").setValue(mstring, "string")
mbool = False
mi11.findParameter("message").setValue(mbool, "bool")
mint = 12672
mi12.findParameter("message").setValue(mint, "int")
mlistdict = [{"SORTIE_@{inparam4}": 4098, "sape_@{inpar2}": 4139}, {"@{inparam4}jj@{inpar2}": 234}]
mi12.findParameter("message").setValue(mlistdict, "list")

sd1.findParameter("result").link("mi5", "result")
# sd1.findParameter('message').link('self','inparam4') # taken from the level of step

w1 = Workflow(name="main")
w1.setOrigin("/home/user/blablabla")
w1.setDescription("Pretty long description\n several lines of text")
w1.setDescrShort("Oooooo short description")
w1.addStep(sd1)

w1.addParameter(Parameter("final", "0.0", "float", "", "", False, True, "Final result"))
w1.addParameter(Parameter("debug", "False", "bool", "", "", True, False, "Debug switch"))
w1.addParameter(Parameter("message", "vv@{inparam4}jj@{inpar2}ge", "string", "", "", True, False, ""))
w1.addParameter(Parameter("inparam4", "VER", "string", "", "", True, False, ""))
w1.addParameter(Parameter("inpar2", "SORTIE@{inparam4}", "string", "", "", True, False, ""))
si1 = w1.createStepInstance("TotalSumm", "si1")
si2 = w1.createStepInstance("TotalSumm", "si2")

si1.findParameter("debug").link("self", "debug")
si2.findParameter("debug").link("self", "debug")
si2.findParameter("input1").link("si1", "result")  # linking the results
w1.findParameter("final").link("si2", "result")

# ============================================================================
# test section
# ============================================================================
w1.toXMLFile("/afs/cern.ch/user/g/gkuznets/test1.xml")
w2 = fromXMLFile("/afs/cern.ch/user/g/gkuznets/test1.xml")
w2.toXMLFile("/afs/cern.ch/user/g/gkuznets/test2.xml")
w4 = fromXMLFile("/afs/cern.ch/user/g/gkuznets/test2.xml")
print(w4.createCode())
eval(compile(w4.createCode(), "<string>", "exec"))
print("===================================================================")
w4.execute()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.DISET.private.Transports import PlainTransport, SSLTransport

gProtocolDict = {
    "dip": {
        "transport": PlainTransport.PlainTransport,
        "sanity": PlainTransport.checkSanity,
        "delegation": PlainTransport.delegate,
    },
    "dips": {
        "transport": SSLTransport.SSLTransport,
        "sanity": SSLTransport.checkSanity,
        "delegation": SSLTransport.delegate,
    },
}

gDefaultProtocol = "dips"



"""
Create and put Requests to archive files.

**List of operations**

#. Optionally replicate files to SourceSE
#. ArchiveFiles: Create a tarball from input files, upload tarball  to TarballSE
#. ReplicateAndRegister Tarball to TargetSE
#. Optionally: Add LFNs to an ArchiveSE
#. Optionally: Check for Tarball Migration
#. Remove all other replicas for these files, or remove all files
#. Remove original replica of Tarball

Will copy all the respective files and place them in to tarballs. Then the tarballs are migrated to
another storage element. Once the file is migrated to tape the original files will be
removed. Optionally the original files can be registered in a special archive SE, so that their
metadata is preserved.

**Related Options**

This script only works if the ``ArchiveFiles`` and ``CheckMigration`` RequestHandlers are configured.
To prevent submission of broken requests the script needs to be enabled in the Operations section of the CS

* Operations/DataManagement/ArchiveFiles/Enabled=True

Default values for any of the command line options can also be set in the CS

* Operations/DataManagement/ArchiveFiles/ArchiveSE
* Operations/DataManagement/ArchiveFiles/TarballSE
* Operations/DataManagement/ArchiveFiles/SourceSE
* Operations/DataManagement/ArchiveFiles/MaxFiles
* ...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os

import DIRAC
from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities import DEncode
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation

sLog = gLogger.getSubLogger("AddArchive")
__RCSID__ = "$Id$"
MAX_SIZE = 2 * 1024 * 1024 * 1024  # 2 GB
MAX_FILES = 2000


class CreateArchiveRequest(object):
    """Create the request to archive files."""

    def __init__(self):
        """Constructor."""
        self._fcClient = None
        self._reqClient = None
        self.switches = {}
        self.requests = []
        self.lfnList = []
        self.metaData = None
        self.options = [
            ("A", "ArchiveSE", "SE for registering archive files at"),
            ("I", "TarballSE", "SE to initially upload tarball"),
            ("P", "Path", "LFN path to folder, all files in the folder will be archived"),
            ("N", "Name", "Name of the Tarball, if not given: Path_Tars/Path_N.tar" " will be used to store tarballs"),
            ("L", "List", "File containing list of LFNs to archive, requires Name to be given"),
            ("", "MaxFiles", "Maximum number to put in one tarball: Default %d" % MAX_FILES),
            ("", "MaxSize", "Maximum number of Bytes to put in one tarball: Default %d" % MAX_SIZE),
            ("S", "SourceSE", "Where to remove the LFNs from"),
            ("T", "TargetSE", "Where to move the Tarball to"),
        ]
        self.flags = [
            ("M", "ReplicateTarball", "Replicate the tarball"),
            ("C", "CheckMigration", "Ensure the tarball is migrated to tape before removing any files or replicas"),
            ("D", "RemoveReplicas", "Remove Replicas from non-ArchiveSE"),
            ("U", "RemoveFiles", "Remove Archived files completely"),
            ("R", "RegisterDescendent", "Register the Tarball as a descendent of the archived LFNs"),
            ("", "AllowReplication", "Enable first replicating to Source-SE"),
            ("", "SourceOnly", "Only treat files that are already at the Source-SE"),
            ("X", "Execute", "Put Requests, else dryrun"),
        ]
        self.registerSwitchesAndParseCommandLine()

        self.switches["MaxSize"] = int(self.switches.setdefault("MaxSize", MAX_SIZE))
        self.switches["MaxFiles"] = int(self.switches.setdefault("MaxFiles", MAX_FILES))

        self.getLFNList()
        self.getLFNMetadata()
        self.lfnChunks = []
        self.replicaSEs = []

    @property
    def fcClient(self):
        """Return FileCatalogClient."""
        if not self._fcClient:
            from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

            self._fcClient = FileCatalog()
        return self._fcClient

    @property
    def reqClient(self):
        """Return RequestClient."""
        if not self._reqClient:
            from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient

            self._reqClient = ReqClient()
        return self._reqClient

    @property
    def dryRun(self):
        """Return dry run flag."""
        return self.switches["DryRun"]

    @property
    def targetSE(self):
        """Return the list of targetSE."""
        return self.switches["TargetSE"]

    @property
    def sourceSEs(self):
        """Return the list of sourceSEs."""
        return self.switches["SourceSE"]

    @property
    def name(self):
        """Return the name of the Request."""
        return self.switches.get("Name", None)

    @property
    def lfnFolderPath(self):
        """Return the lfn folder path where to find the files of the request."""
        return self.switches.get("Path", None)

    def registerSwitchesAndParseCommandLine(self):
        """Register the default plus additional parameters and parse options.

        :param list options: list of three tuple for options to add to the script
        :param list flags:  list of three tuple for flags to add to the script
        :param str opName
        """
        for short, longOption, doc in self.options:
            Script.registerSwitch(short + ":" if short else "", longOption + "=", doc)
        for short, longOption, doc in self.flags:
            Script.registerSwitch(short, longOption, doc)
            self.switches[longOption] = False
        Script.parseCommandLine()
        if Script.getPositionalArgs():
            Script.showHelp(exitCode=1)

        ops = Operations()
        if not ops.getValue("DataManagement/ArchiveFiles/Enabled", False):
            sLog.error('The "ArchiveFiles" operation is not enabled, contact your administrator!')
            DIRAC.exit(1)
        for _short, longOption, _doc in self.options:
            defaultValue = ops.getValue("DataManagement/ArchiveFiles/%s" % longOption, None)
            if defaultValue:
                sLog.verbose("Found default value in the CS for %r with value %r" % (longOption, defaultValue))
                self.switches[longOption] = defaultValue
        for _short, longOption, _doc in self.flags:
            defaultValue = ops.getValue("DataManagement/ArchiveFiles/%s" % longOption, False)
            if defaultValue:
                sLog.verbose("Found default value in the CS for %r with value %r" % (longOption, defaultValue))
                self.switches[longOption] = defaultValue

        for switch in Script.getUnprocessedSwitches():
            for short, longOption, doc in self.options:
                if switch[0] == short or switch[0].lower() == longOption.lower():
                    sLog.verbose("Found switch %r with value %r" % (longOption, switch[1]))
                    self.switches[longOption] = switch[1]
                    break
            for short, longOption, doc in self.flags:
                if switch[0] == short or switch[0].lower() == longOption.lower():
                    self.switches[longOption] = True
                    break

        self.checkSwitches()
        self.switches["DryRun"] = not self.switches.get("Execute", False)
        self.switches["SourceSE"] = self.switches.get("SourceSE", "").split(",")

    def getLFNList(self):
        """Get list of LFNs.

        Either read the provided file, or get the files found beneath the provided folder.

        :param dict switches: options from command line
        :returns: list of lfns
        :raises: RuntimeError, ValueError
        """
        if self.switches.get("List"):
            if os.path.exists(self.switches.get("List")):
                self.lfnList = list(
                    set([line.split()[0] for line in open(self.switches.get("List")).read().splitlines()])
                )
            else:
                raise ValueError("%s not a file" % self.switches.get("List"))
        elif self.lfnFolderPath:
            path = self.lfnFolderPath
            sLog.debug("Check if %r is a directory" % path)
            isDir = returnSingleResult(self.fcClient.isDirectory(path))
            sLog.debug("Result: %r" % isDir)
            if not isDir["OK"] or not isDir["Value"]:
                sLog.error("Path is not a directory", isDir.get("Message", ""))
                raise RuntimeError("Path %r is not a directory" % path)
            sLog.notice("Looking for files in %r" % path)

            metaDict = {"SE": self.sourceSEs[0]} if self.switches.get("SourceOnly") else {}
            lfns = self.fcClient.findFilesByMetadata(metaDict=metaDict, path=path)
            if not lfns["OK"]:
                sLog.error("Could not find files")
                raise RuntimeError(lfns["Message"])
            self.lfnList = lfns["Value"]

        if self.lfnList:
            sLog.notice("Will create request(s) with %d lfns" % len(self.lfnList))
            if len(self.lfnList) == 1:
                raise RuntimeError("Only 1 file in the list, aborting!")
            return

        raise ValueError('"Path" or "List" need to be provided!')

    def putOrRunRequests(self):
        """Run or put requests."""
        requestIDs = []

        if self.dryRun:
            sLog.notice("Would have created %d requests" % len(self.requests))
            for reqID, req in enumerate(self.requests):
                sLog.notice("Request %d:" % reqID)
                for opID, op in enumerate(req):
                    sLog.notice("        Operation %d: %s #lfn %d" % (opID, op.Type, len(op)))
            return 0
        for request in self.requests:
            putRequest = self.reqClient.putRequest(request)
            if not putRequest["OK"]:
                sLog.error("unable to put request %r: %s" % (request.RequestName, putRequest["Message"]))
                continue
            requestIDs.append(str(putRequest["Value"]))
            sLog.always("Request %r has been put to ReqDB for execution." % request.RequestName)

        if requestIDs:
            sLog.always("%d requests have been put to ReqDB for execution" % len(requestIDs))
            sLog.always("RequestID(s): %s" % " ".join(requestIDs))
            sLog.always("You can monitor the request status using the command: dirac-rms-request <requestName/ID>")
            return 0

        sLog.error("No requests created")
        return 1

    def checkSwitches(self):
        """Check the switches, set autoName if needed."""
        if not self.switches.get("SourceSE"):
            raise RuntimeError('Have to set "SourceSE"')
        if not self.switches.get("List") and not self.switches.get("Path"):
            raise RuntimeError('Have to set "List" or "Path"')
        if not self.name and self.lfnFolderPath:
            self.switches["AutoName"] = os.path.join(
                os.path.dirname(self.lfnFolderPath), os.path.basename(self.lfnFolderPath) + ".tar"
            )
            sLog.notice("Using %r for tarball" % self.switches.get("AutoName"))

        if self.switches.get("List") and not self.name:
            raise RuntimeError('Have to set "Name" with "List"')

        if self.switches.get("RemoveReplicas") and self.switches.get("ArchiveSE") is None:
            sLog.error("'RemoveReplicas' does not work without 'ArchiveSE'")
            raise RuntimeError("ArchiveSE missing")

        if self.switches.get("RemoveReplicas") and self.switches.get("RemoveFiles"):
            sLog.error("Use either 'RemoveReplicas' or 'RemoveFiles', not both!")
            raise RuntimeError("Too many removal flags")

        if self.switches.get("ReplicateTarball") and not self.switches.get("TargetSE"):
            sLog.error("Have to set 'TargetSE' with 'ReplicateTarball'")
            raise RuntimeError("ReplicateTarball missing TargetSE")

    def splitLFNsBySize(self):
        """Split LFNs into MAX_SIZE chunks of at most MAX_FILES length.

        :return: list of list of lfns
        """
        sLog.notice("Splitting files by Size")
        lfnChunk = []
        totalSize = 0
        for lfn, info in self.metaData["Successful"].items():
            if totalSize > self.switches["MaxSize"] or len(lfnChunk) >= self.switches["MaxFiles"]:
                self.lfnChunks.append(lfnChunk)
                sLog.notice("Created Chunk of %s lfns with %s bytes" % (len(lfnChunk), totalSize))
                lfnChunk = []
                totalSize = 0
            lfnChunk.append(lfn)
            totalSize += info["Size"]

        self.lfnChunks.append(lfnChunk)
        sLog.notice("Created Chunk of %s lfns with %s bytes" % (len(lfnChunk), totalSize))

        self.replicaSEs = set(
            [
                seItem
                for se in self.fcClient.getReplicas(self.lfnList)["Value"]["Successful"].values()
                for seItem in se.keys()
            ]
        )

    def run(self):
        """Perform checks and create the request."""
        if self.switches.get("AutoName"):
            baseArchiveLFN = archiveLFN = self.switches["AutoName"]
            tarballName = os.path.basename(archiveLFN)
        else:
            baseArchiveLFN = archiveLFN = self.name
            tarballName = os.path.basename(archiveLFN)
        baseRequestName = requestName = "Archive_%s" % tarballName.rsplit(".", 1)[0]

        from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator

        self.splitLFNsBySize()

        for count, lfnChunk in enumerate(self.lfnChunks):
            if not lfnChunk:
                sLog.error("LFN list is empty!!!")
                return 1

            if len(self.lfnChunks) > 1:
                requestName = "%s_%d" % (baseRequestName, count)
                baseName = os.path.split(baseArchiveLFN.rsplit(".", 1)[0])
                archiveLFN = "%s/%s_Tars/%s_%d.tar" % (baseName[0], baseName[1], baseName[1], count)

            self.checkArchive(archiveLFN)

            request = self.createRequest(requestName, archiveLFN, lfnChunk)

            valid = RequestValidator().validate(request)
            if not valid["OK"]:
                sLog.error("putRequest: request not valid", "%s" % valid["Message"])
                return 1
            else:
                self.requests.append(request)

        self.putOrRunRequests()
        return 0

    def addLFNs(self, operation, lfns, addPFN=False):
        """Add lfns to operation.

        :param operation: the operation instance to which the files will be added
        :param list lfns: list of lfns
        :param bool addPFN: if true adds PFN to each File
        """
        if not self.metaData:
            self.getLFNMetadata()

        for lfn in lfns:
            metaDict = self.metaData["Successful"][lfn]
            opFile = File()
            opFile.LFN = lfn
            if addPFN:
                opFile.PFN = lfn
            opFile.Size = metaDict["Size"]
            if "Checksum" in metaDict:
                # should check checksum type, now assuming Adler32 (metaDict['ChecksumType'] = 'AD')
                opFile.Checksum = metaDict["Checksum"]
                opFile.ChecksumType = "ADLER32"
            operation.addFile(opFile)

    def getLFNMetadata(self):
        """Get the metadata for all the LFNs."""
        metaData = self.fcClient.getFileMetadata(self.lfnList)
        error = False
        if not metaData["OK"]:
            sLog.error("Unable to read metadata for lfns: %s" % metaData["Message"])
            raise RuntimeError("Could not read metadata: %s" % metaData["Message"])

        self.metaData = metaData["Value"]
        for failedLFN, reason in self.metaData["Failed"].items():
            sLog.error("skipping %s: %s" % (failedLFN, reason))
            error = True
        if error:
            raise RuntimeError("Could not read all metadata")

        for lfn in self.metaData["Successful"].keys():
            sLog.verbose("found %s" % lfn)

    def createRequest(self, requestName, archiveLFN, lfnChunk):
        """Create the Request."""
        request = Request()
        request.RequestName = requestName

        self._checkReplicaSites(request, lfnChunk)

        archiveFiles = Operation()
        archiveFiles.Type = "ArchiveFiles"
        archiveFiles.Arguments = DEncode.encode(
            {
                "SourceSE": self.sourceSEs[0],
                "TarballSE": self.switches["TarballSE"],
                "RegisterDescendent": self.switches["RegisterDescendent"],
                "ArchiveLFN": archiveLFN,
            }
        )
        self.addLFNs(archiveFiles, lfnChunk)
        request.addOperation(archiveFiles)

        # Replicate the Tarball, ArchiveFiles will upload it
        if self.switches.get("ReplicateTarball"):
            replicateAndRegisterTarBall = Operation()
            replicateAndRegisterTarBall.Type = "ReplicateAndRegister"
            replicateAndRegisterTarBall.TargetSE = self.targetSE
            opFile = File()
            opFile.LFN = archiveLFN
            replicateAndRegisterTarBall.addFile(opFile)
            request.addOperation(replicateAndRegisterTarBall)

        if self.switches.get("CheckMigration"):
            checkMigrationTarBall = Operation()
            checkMigrationTarBall.Type = "CheckMigration"
            migrationTarget = self.targetSE if self.switches.get("ReplicateTarball") else self.switches["TarballSE"]
            checkMigrationTarBall.TargetSE = migrationTarget
            opFile = File()
            opFile.LFN = archiveLFN
            checkMigrationTarBall.addFile(opFile)
            request.addOperation(checkMigrationTarBall)

        # Register Archive Replica for LFNs
        if self.switches.get("ArchiveSE"):
            registerArchived = Operation()
            registerArchived.Type = "RegisterReplica"
            registerArchived.TargetSE = self.switches.get("ArchiveSE")
            self.addLFNs(registerArchived, lfnChunk, addPFN=True)
            request.addOperation(registerArchived)

            # Remove all Other Replicas for LFNs
            if self.switches.get("RemoveReplicas"):
                removeArchiveReplicas = Operation()
                removeArchiveReplicas.Type = "RemoveReplica"
                removeArchiveReplicas.TargetSE = ",".join(self.replicaSEs)
                self.addLFNs(removeArchiveReplicas, lfnChunk)
                request.addOperation(removeArchiveReplicas)

        # Remove all Replicas for LFNs
        if self.switches.get("RemoveFiles"):
            removeArchiveFiles = Operation()
            removeArchiveFiles.Type = "RemoveFile"
            self.addLFNs(removeArchiveFiles, lfnChunk)
            request.addOperation(removeArchiveFiles)

        # Remove Original tarball replica
        if self.switches.get("ReplicateTarball"):
            removeTarballOrg = Operation()
            removeTarballOrg.Type = "RemoveReplica"
            removeTarballOrg.TargetSE = self.sourceSEs[0]
            opFile = File()
            opFile.LFN = archiveLFN
            removeTarballOrg.addFile(opFile)
            request.addOperation(removeTarballOrg)
        return request

    def checkArchive(self, archiveLFN):
        """Check that archiveLFN does not exist yet."""
        sLog.notice("Using Tarball: %s" % archiveLFN)
        exists = returnSingleResult(self.fcClient.isFile(archiveLFN))
        sLog.debug("Checking for Tarball existence %r" % exists)
        if exists["OK"] and exists["Value"]:
            raise RuntimeError("Tarball %r already exists" % archiveLFN)

        sLog.debug("Checking permissions for %r" % archiveLFN)
        hasAccess = returnSingleResult(self.fcClient.hasAccess(archiveLFN, "addFile"))
        if not archiveLFN or not hasAccess["OK"] or not hasAccess["Value"]:
            sLog.error("Error checking tarball location: %r" % hasAccess)
            raise ValueError('%s is not a valid path, parameter "Name" must be correct' % archiveLFN)

    def _checkReplicaSites(self, request, lfnChunk):
        """Ensure that all lfns can be found at the SourceSE, otherwise add replication operation to request.

        If SourceOnly is set just rejetct those LFNs.

        """
        resReplica = self.fcClient.getReplicas(lfnChunk)
        if not resReplica["OK"]:
            sLog.error("Failed to get replica information:", resReplica["Message"])
            raise RuntimeError("Failed to get replica information")

        atSource = []
        notAt = []
        failed = []
        sourceSE = self.sourceSEs[0]
        for lfn, replInfo in resReplica["Value"]["Successful"].items():
            if sourceSE in replInfo:
                atSource.append(lfn)
            else:
                sLog.notice("WARN: LFN %r not found at source, only at: %s" % (lfn, ",".join(replInfo.keys())))
                notAt.append(lfn)

        for lfn, errorMessage in resReplica["Value"]["Failed"].items():
            sLog.error("Failed to get replica info", "%s: %s" % (lfn, errorMessage))
            failed.append(lfn)

        if failed:
            raise RuntimeError("Failed to get replica information")
        sLog.notice("Found %d files to replicate" % len(notAt))
        if not notAt:
            return
        if notAt and self.switches.get("AllowReplication"):
            self._replicateSourceFiles(request, notAt)
        else:
            raise RuntimeError("Not all files are at the Source, exiting")

    def _replicateSourceFiles(self, request, lfns):
        """Create the replicateAndRegisterRequest.

        :param request: The request to add the operation to
        :param lfns: list of LFNs
        """
        registerSource = Operation()
        registerSource.Type = "ReplicateAndRegister"
        registerSource.TargetSE = self.sourceSEs[0]
        self.addLFNs(registerSource, lfns, addPFN=True)
        request.addOperation(registerSource)


@Script()
def main():
    try:
        CAR = CreateArchiveRequest()
        CAR.run()
    except Exception as e:
        if LogLevels.getLevelValue(sLog.getLevel()) <= LogLevels.VERBOSE:
            sLog.exception("Failed to create Archive Request")
        else:
            sLog.error("ERROR: Failed to create Archive Request:", str(e))
        exit(1)
    exit(0)


if __name__ == "__main__":
    main()



""" DISET request handler base class for the ProductionDB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

prodTypes = [six.string_types, int]
transTypes = [six.string_types, int, list]


class ProductionManagerHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB object"""

        try:
            result = ObjectLoader().loadObject("ProductionSystem.DB.ProductionDB", "ProductionDB")
            if not result["OK"]:
                return result
            cls.productionDB = result["Value"]()
        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        return S_OK()

    ####################################################################
    #
    # These are the methods to manipulate the Productions table
    #

    types_addProduction = [six.string_types, six.string_types]

    def export_addProduction(self, prodName, prodDescription):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        authorGroup = credDict.get("group")
        res = self.productionDB.addProduction(prodName, prodDescription, authorDN, authorGroup)
        if res["OK"]:
            gLogger.info("Added production %d" % res["Value"])
        return res

    types_deleteProduction = [prodTypes]

    def export_deleteProduction(self, prodName):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        return self.productionDB.deleteProduction(prodName, author=authorDN)

    types_getProductions = []

    @classmethod
    def export_getProductions(
        cls,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="CreationDate",
        orderAttribute=None,
        limit=None,
        offset=None,
    ):
        if not condDict:
            condDict = {}
        return cls.productionDB.getProductions(
            condDict=condDict,
            older=older,
            newer=newer,
            timeStamp=timeStamp,
            orderAttribute=orderAttribute,
            limit=limit,
            offset=offset,
        )

    types_getProduction = [prodTypes]

    @classmethod
    def export_getProduction(cls, prodName):
        return cls.productionDB.getProduction(prodName)

    types_getProductionParameters = [prodTypes, [six.string_types, list, tuple]]

    @classmethod
    def export_getProductionParameters(cls, prodName, parameters):
        return cls.productionDB.getProductionParameters(prodName, parameters)

    types_setProductionStatus = [prodTypes, six.string_types]

    @classmethod
    def export_setProductionStatus(cls, prodName, status):
        return cls.productionDB.setProductionStatus(prodName, status)

    types_startProduction = [prodTypes]

    @classmethod
    @ignoreEncodeWarning
    def export_startProduction(cls, prodName):
        return cls.productionDB.startProduction(prodName)

    ####################################################################
    #
    # These are the methods to manipulate the ProductionTransformations table
    #

    types_addTransformationsToProduction = [prodTypes, transTypes, transTypes]

    @classmethod
    def export_addTransformationsToProduction(cls, prodName, transIDs, parentTransIDs):
        return cls.productionDB.addTransformationsToProduction(prodName, transIDs, parentTransIDs=parentTransIDs)

    types_getProductionTransformations = []

    @classmethod
    def export_getProductionTransformations(
        cls,
        prodName,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="CreationTime",
        orderAttribute=None,
        limit=None,
        offset=None,
    ):

        if not condDict:
            condDict = {}
        return cls.productionDB.getProductionTransformations(
            prodName,
            condDict=condDict,
            older=older,
            newer=newer,
            timeStamp=timeStamp,
            orderAttribute=orderAttribute,
            limit=limit,
            offset=offset,
        )

    ####################################################################
    #
    # These are the methods to manipulate the ProductionSteps table
    #

    types_addProductionStep = [dict]

    @classmethod
    def export_addProductionStep(cls, prodStep):
        stepName = prodStep["name"]
        stepDescription = prodStep["description"]
        stepLongDescription = prodStep["longDescription"]
        stepBody = prodStep["body"]
        stepType = prodStep["stepType"]
        stepPlugin = prodStep["plugin"]
        stepAgentType = prodStep["agentType"]
        stepGroupSize = prodStep["groupsize"]
        stepInputQuery = prodStep["inputquery"]
        stepOutputQuery = prodStep["outputquery"]
        res = cls.productionDB.addProductionStep(
            stepName,
            stepDescription,
            stepLongDescription,
            stepBody,
            stepType,
            stepPlugin,
            stepAgentType,
            stepGroupSize,
            stepInputQuery,
            stepOutputQuery,
        )
        if res["OK"]:
            gLogger.info("Added production step %d" % res["Value"])
        return res

    types_getProductionStep = [int]

    @classmethod
    def export_getProductionStep(cls, stepID):
        return cls.productionDB.getProductionStep(stepID)



""" Test class for plugins
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name,
# line-too-long

# imports
import pytest
from mock import MagicMock

from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock

from DIRAC import gLogger, S_OK

from DIRAC.TransformationSystem.Agent.TransformationPlugin import TransformationPlugin

paramsBase = {
    "AgentType": "Automatic",
    "DerivedProduction": "0",
    "FileMask": "",
    "GroupSize": 1,
    "InheritedFrom": 0,
    "JobType": "MCSimulation",
    "MaxNumberOfTasks": 0,
    "OutputDirectories": "['/lhcb/MC/20', '/lhcb/debug/20']",
    "OutputLFNs": "{'LogTargetPath': ['/lhcb/9.tar'], 'LogFilePath': ['/lhcb/9']}",
    "Priority": "0",
    "SizeGroup": "1",
    "Status": "Active",
    "TransformationID": 1080,
    "Type": "MCSimulation",
    "outputDataFileMask": "GAUSSHIST;ALLSTREAMS.DST",
}

data = {
    "/this/is/at.1": ["SE1"],
    "/this/is/at.2": ["SE2"],
    "/this/is/als/at.2": ["SE2"],
    "/this/is/at.12": ["SE1", "SE2"],
    "/this/is/also/at.12": ["SE1", "SE2"],
    "/this/is/at_123": ["SE1", "SE2", "SE3"],
    "/this/is/at_23": ["SE2", "SE3"],
    "/this/is/at_4": ["SE4"],
}


@pytest.fixture
def setup(mocker):
    tpName = "DIRAC.TransformationSystem.Agent.TransformationPlugin"
    tuName = "DIRAC.TransformationSystem.Client.Utilities"
    mocker.patch(tpName + ".TransformationClient", new=MagicMock("tClient"))
    mocker.patch(tpName + ".TransformationPlugin._getSiteForSE", return_value=S_OK([]))
    mocker.patch(tuName + ".DataManager", new=dm_mock)
    mocker.patch(tuName + ".FileCatalog", new=fc_mock)
    mocker.patch(tuName + ".StorageElement", new=MagicMock("SEMock"))
    gLogger.setLevel("DEBUG")


def test__Standard_G10(setup):
    """Test StandardPlugin: no input data, active."""
    params = dict(paramsBase)
    params["GroupSize"] = 10
    pluginStandard = TransformationPlugin("Standard")
    pluginStandard.setParameters(params)
    res = pluginStandard.run()
    assert res["OK"]
    # no data
    assert res["Value"] == []


def test__Standard_Data_G10(setup):
    """Test StandardPlugin: input data, active."""
    params = dict(paramsBase)
    params["GroupSize"] = 10
    pluginStandard = TransformationPlugin("Standard")
    pluginStandard.setParameters(params)
    pluginStandard.setInputData(data)
    res = pluginStandard.run()
    assert res["OK"]
    # data less than 10
    assert res["Value"] == []


def test__Standard_Flush_G10(setup):
    """Test StandardPlugin: input data, flush."""
    pluginStandard = TransformationPlugin("Standard")
    params = dict(paramsBase)
    params["GroupSize"] = 10
    params["Status"] = "Flush"
    pluginStandard.setParameters(params)
    pluginStandard.setInputData(data)
    res = pluginStandard.run()
    sortedData = [
        ("SE1", {"/this/is/at.1"}),
        ("SE1,SE2", {"/this/is/also/at.12", "/this/is/at.12"}),
        ("SE1,SE2,SE3", {"/this/is/at_123"}),
        ("SE2", {"/this/is/als/at.2", "/this/is/at.2"}),
        ("SE2,SE3", {"/this/is/at_23"}),
        ("SE4", {"/this/is/at_4"}),
    ]
    assert res["OK"]
    assert [(a, set(b)) for a, b in res["Value"]] == sortedData


def test__Standard_G1(setup):
    """Test StandardPlugin: not input data, active."""
    pluginStandard = TransformationPlugin("Standard")
    pluginStandard.setParameters(paramsBase)
    res = pluginStandard.run()
    assert res["OK"]
    assert res["Value"] == []


def test__Standard_Data_G1(setup):
    """Test StandardPlugin: input data, active."""
    pluginStandard = TransformationPlugin("Standard")
    pluginStandard.setParameters(paramsBase)
    pluginStandard.setInputData(data)
    res = pluginStandard.run()
    assert res["OK"]
    sortedData = sorted([(",".join(SEs), [lfn]) for lfn, SEs in data.items()])
    assert res["Value"], sortedData


def test__Standard_Flush_G1(setup):
    """Test StandardPlugin: input data, flush."""
    pluginStandard = TransformationPlugin("Standard")
    params = dict(paramsBase)
    params["Status"] = "Flush"
    pluginStandard.setParameters(params)
    pluginStandard.setInputData(data)
    res = pluginStandard.run()
    sortedData = sorted([(",".join(SEs), [lfn]) for lfn, SEs in data.items()])
    assert res["OK"]
    assert sorted(res["Value"]) == sorted(sortedData)


def test__Broadcast_Active_G1(setup):
    """Test BroadcastPlugin: input data, Active"""
    thePlugin = TransformationPlugin("Broadcast")
    params = dict(paramsBase)
    params["Status"] = "Active"
    params["SourceSE"] = "SE1"
    params["TargetSE"] = "SE2"
    thePlugin.setParameters(params)
    thePlugin.setInputData(data)
    res = thePlugin.run()
    # sort returned values by first lfn in LFNs
    sortedReturn = [(SE, lfns) for SE, lfns in sorted(res["Value"], key=lambda t: t[1][0])]
    # sort data by lfn
    expected = [("SE2", [lfn]) for lfn, SEs in sorted(data.items(), key=lambda t: t[0]) if "SE1" in SEs]
    assert res["OK"]
    assert sortedReturn == expected


def test__Broadcast_Active_G10(setup):
    """Test BroadcastPlugin: input data, Active"""
    thePlugin = TransformationPlugin("Broadcast")
    params = dict(paramsBase)
    params["Status"] = "Active"
    params["SourceSE"] = "['SE1']"
    params["TargetSE"] = ["SE2"]
    params["GroupSize"] = 10
    thePlugin.setParameters(params)
    thePlugin.setInputData(data)
    res = thePlugin.run()
    assert res["OK"]
    assert res["Value"] == []


def test__Broadcast_Active_G1_NoSource(setup):
    """Test BroadcastPlugin: input data, Active, noSource"""
    thePlugin = TransformationPlugin("Broadcast")
    params = dict(paramsBase)
    params["Status"] = "Active"
    params["TargetSE"] = ["SE2"]
    params["GroupSize"] = 1
    thePlugin.setParameters(params)
    thePlugin.setInputData(data)
    res = thePlugin.run()
    # sort returned values by first lfn in LFNs
    sortedReturn = [(SE, lfns) for SE, lfns in sorted(res["Value"], key=lambda t: t[1][0])]
    # sort data by lfn
    expected = [("SE2", [lfn]) for lfn, _SEs in sorted(data.items(), key=lambda t: t[0])]
    assert res["OK"]
    assert sortedReturn == expected



#!/usr/bin/env python
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    fieldsToShow = ("ComponentName", "Type", "Host", "Port", "Status", "Message")

    from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor

    result = gMonitor.getComponentsStatusWebFormatted(sortingList=[["ComponentName", "ASC"]])
    if not result["OK"]:
        print("ERROR: %s" % result["Message"])
        sys.exit(1)
    paramNames = result["Value"]["ParameterNames"]
    records = result["Value"]["Records"]
    fieldLengths = []
    for param in paramNames:
        fieldLengths.append(len(param))

    for record in records:
        for i, _ in enumerate(record):
            if paramNames[i] in fieldsToShow:
                fieldLengths[i] = max(fieldLengths[i], len(str(record[i])))
    # Print time!
    line = []
    sepLine = []
    for i, param in enumerate(paramNames):
        if param in fieldsToShow:
            line.append("%s%s" % (param, " " * (fieldLengths[i] - len(param))))
            sepLine.append("-" * fieldLengths[i])
    print("|".join(line))
    sepLine = "+".join(sepLine)
    print(sepLine)
    for record in records:
        line = []
        for i, _ in enumerate(record):
            if paramNames[i] in fieldsToShow:
                val = str(record[i])
                line.append("%s%s" % (val, " " * (fieldLengths[i] - len(val))))
        print("|".join(line))
        # print sepLine


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.Utilities import DEncode


class UserProfileClient(object):
    def __init__(self, profile, rpcClientFunctor=RPCClient):
        self.rpcClientFunctor = rpcClientFunctor
        self.profile = profile

    def __getRPCClient(self):
        return self.rpcClientFunctor("Framework/UserProfileManager")

    def storeVar(self, varName, data, perms={}):
        try:
            stub = DEncode.encode(data)
        except Exception as e:
            return S_ERROR("Cannot encode data:%s" % str(e))
        return self.__getRPCClient().storeProfileVar(self.profile, varName, stub, perms)

    def __decodeVar(self, data):
        try:
            dataObj, lenData = DEncode.decode(data.encode())
        except Exception as e:
            return S_ERROR("Cannot decode data: %s" % str(e))
        return S_OK(dataObj)

    def retrieveVar(self, varName):
        rpcClient = self.__getRPCClient()
        result = rpcClient.retrieveProfileVar(self.profile, varName)
        if not result["OK"]:
            return result
        return self.__decodeVar(result["Value"])

    def retrieveVarFromUser(self, ownerName, ownerGroup, varName):
        rpcClient = self.__getRPCClient()
        result = rpcClient.retrieveProfileVarFromUser(ownerName, ownerGroup, self.profile, varName)
        if not result["OK"]:
            return result
        return self.__decodeVar(result["Value"])

    def retrieveAllVars(self):
        rpcClient = self.__getRPCClient()
        result = rpcClient.retrieveProfileAllVars(self.profile)
        if not result["OK"]:
            return result
        try:
            encodedData = result["Value"]
            dataObj = {key: DEncode.decode(value.encode())[0] for key, value in encodedData.items()}
        except Exception as e:
            return S_ERROR("Cannot decode data: %s" % str(e))
        return S_OK(dataObj)

    def listAvailableVars(self, filterDict={}):
        rpcClient = self.__getRPCClient()
        return rpcClient.listAvailableProfileVars(self.profile, filterDict)

    def getUserProfiles(self):
        rpcClient = self.__getRPCClient()
        return rpcClient.getUserProfiles()

    def setVarPermissions(self, varName, perms):
        rpcClient = self.__getRPCClient()
        return rpcClient.setProfileVarPermissions(self.profile, varName, perms)

    def getVarPermissions(self, varName):
        rpcClient = self.__getRPCClient()
        return rpcClient.getProfileVarPermissions(self.profile, varName)

    def deleteVar(self, varName):
        return self.__getRPCClient().deleteProfileVar(self.profile, varName)

    def deleteProfiles(self, userList):
        return self.__getRPCClient().deleteProfiles(userList)

    def getUserProfileNames(self, permission=dict()):
        """
        it returns the available profile names by not taking account the permission: ReadAccess and PublishAccess
        """
        return self.__getRPCClient().getUserProfileNames(permission)

    def listStatesForWeb(self, permission={}):
        return self.__getRPCClient().listStatesForWeb(permission)



""" DIRAC FileCatalog Security Manager mix-in class with no access checks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.SecurityManager.SecurityManagerBase import SecurityManagerBase


class NoSecurityManager(SecurityManagerBase):
    def getPathPermissions(self, paths, credDict):
        """Get path permissions according to the policy"""

        permissions = {}
        for path in paths:
            permissions[path] = {"Read": True, "Write": True, "Execute": True}

        return S_OK({"Successful": permissions, "Failed": {}})

    def hasAccess(self, opType, paths, credDict):
        successful = {}
        for path in paths:
            successful[path] = True
        resDict = {"Successful": successful, "Failed": {}}
        return S_OK(resDict)

    def hasAdminAccess(self, credDict):
        return S_OK(True)
